# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A Config that builds a `dict` and accepts any parameters."""

from fiddle import config


def _kwargs_to_dict(**kwargs):
  return dict(kwargs)


class DictConfig(config.Config):
  """A Config instance that builds a `dict` accepting any keys/values."""

  def __init__(self, *args, **kwargs):
    super().__init__(_kwargs_to_dict, *args, **kwargs)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to output representations of `fdl.Buildable`s."""

import dataclasses
import inspect
from typing import Any, Iterator, List, Optional, Sequence, Tuple, Type, Union

from fiddle import config
from fiddle import history
from fiddle import tagging
from fiddle.codegen import formatting_utilities
import tree


@dataclasses.dataclass(frozen=True)
class _ParamName:
  """Wrapper for a string to differentiate Buildable param names from dict keys."""
  __slots__ = ('name',)
  name: str


@dataclasses.dataclass(frozen=True)
class _UnsetValue:
  """A wrapper class indicating an unset value."""
  __slots__ = ('parameter',)
  parameter: inspect.Parameter

  def __repr__(self) -> str:
    if self.parameter.default is self.parameter.empty:
      return '<[unset]>'
    else:
      default_value = _format_value(
          self.parameter.default, raw_value_repr=False)
      return f'<[unset; default: {default_value}]>'


_Leaf = Any
_Path = Sequence[Union[int, str, _ParamName]]


def _path_to_str(path: _Path) -> str:
  """Converts a path to a string representation."""
  output = ''
  for item in path:
    if isinstance(item, _ParamName):
      if output:
        output += '.' + item.name
      else:
        # First item.
        output = item.name
    else:
      output += f'[{item!r}]'
  return output


def _has_nested_builder(children: Sequence[Tuple[_Path, _Leaf]]) -> bool:
  return any(
      isinstance(child, config.Buildable) for unused_path, child in children)


def _format_value(value: Any, *, raw_value_repr: bool) -> str:
  if raw_value_repr:
    return repr(value)
  else:
    if isinstance(value, str):
      return repr(value)
    else:
      return formatting_utilities.pretty_print(value)


@dataclasses.dataclass(frozen=True)
class _TaggedValueWrapper:
  """Customizes representation for TaggedValues in flattened output."""
  __slots__ = ('wrapped',)
  wrapped: tagging.TaggedValue

  def __repr__(self):
    if self.wrapped.value is tagging.NO_VALUE:
      value_repr = '<[unset]>'
    else:
      value_repr = repr(self.wrapped.value)
    tag_str = ' '.join(sorted(str(t) for t in self.wrapped.tags))
    return f'{value_repr} {tag_str}'


def as_str_flattened(cfg: config.Buildable,
                     *,
                     include_types: bool = True,
                     raw_value_repr: bool = False) -> str:
  """Returns a string of cfg's paths and values, one pair per line.

  Some automated tools (e.g. grep, sort, diff, ...) handle data line-by-line.
  This output format contains the full path to a value and a string
  representation of said value on a single line.

  Args:
    cfg: A buildable to generate a string representation for.
    include_types: If true, include type annotations (where available) in the
      string representation of each leaf value.
    raw_value_repr: If true, use `repr` on values, otherwise, a custom
      pretty-printed format is used.
  Returns: a string representation of `cfg`.
  """

  def flatten_children(
      value: Any, annotation: Optional[Type[Any]],
      path: _Path) -> Iterator[Tuple[_Path, Optional[Type[Any]], _Leaf]]:
    flattened_children = tree.flatten_with_path(value)
    if _has_nested_builder(flattened_children):
      for child_path, leaf in flattened_children:
        if isinstance(leaf, tagging.TaggedValue):
          yield path + child_path, None, _TaggedValueWrapper(leaf)
        elif isinstance(leaf, config.Buildable):
          for subpath, subannotation, actual_leaf in recursive_flatten(leaf):
            yield path + child_path + subpath, subannotation, actual_leaf
        else:
          yield path + child_path, None, leaf
    else:
      yield path, annotation, value

  def recursive_flatten(
      buildable: config.Buildable
  ) -> Iterator[Tuple[_Path, Optional[Type[Any]], _Leaf]]:
    signature = buildable.__signature__
    kwarg_param = None
    for param_name, param in signature.parameters.items():
      if param.kind == param.VAR_KEYWORD:
        kwarg_param = param
        continue
      annotation = None if param.annotation is param.empty else param.annotation
      path = (_ParamName(param_name),)
      if param_name not in buildable.__arguments__:
        yield path, annotation, _UnsetValue(param)
      else:
        yield from flatten_children(buildable.__arguments__[param_name],
                                    annotation, path)
    kwarg_annotation = None
    if kwarg_param and kwarg_param.annotation is not kwarg_param.empty:
      kwarg_annotation = kwarg_param.annotation
    not_yielded_argument_names = (
        buildable.__arguments__.keys() - signature.parameters.keys())
    for name in sorted(not_yielded_argument_names):
      path = (_ParamName(name),)
      yield from flatten_children(buildable.__arguments__[name],
                                  kwarg_annotation, path)

  def format_line(line: Tuple[_Path, Optional[Type[Any]], _Leaf]):
    type_annotation = ''
    if include_types and line[1] is not None:
      try:
        type_annotation = f': {line[1].__qualname__}'
      except AttributeError:
        # Certain types, such as Union, do not have a __qualname__ attribute.
        type_annotation = f': {line[1]}'
    value = _format_value(line[2], raw_value_repr=raw_value_repr)
    return f'{_path_to_str(line[0])}{type_annotation} = {value}'

  return '\n'.join(map(format_line, recursive_flatten(cfg)))


def history_per_leaf_parameter(cfg: config.Buildable,
                               *,
                               raw_value_repr: bool = False) -> str:
  """Returns a string representing the history of cfg's leaf params.

  Because Buildable's are designed to be mutated, tracking down when and where
  a particular parameter's value is changed can be difficult. This function
  returns a string representation of each parameter and (in
  reverse-chronological order) its current and past values.

  This representation elides the "DAG"-construction aspects of constructing the
  `cfg`, and instead only prints the history of the "outer-most" parameters
  (ones that don't contain a reference to another Buildable).

  Args:
    cfg: A buildable to generate a history for.
    raw_value_repr: If true, use `repr` when string-ifying values, otherwise use
      a customized pretty-printing routine.

  Returns:
    A string representation of `cfg`'s history, organized by param name.
  """
  return '\n'.join(
      _make_per_leaf_histories_recursive(cfg, raw_value_repr=raw_value_repr))


def _make_per_leaf_histories_recursive(
    cfg: config.Buildable,
    raw_value_repr: bool,
    path: Optional[_Path] = None,
) -> Iterator[str]:
  """Recursively traverses `cfg` and generates per-param history summaries."""
  for name, param_history in cfg.__argument_history__.items():
    if path:
      param_path = tuple(path) + (_ParamName(name),)
    else:
      param_path = (_ParamName(name),)
    children = tree.flatten_with_path(getattr(cfg, name))
    if _has_nested_builder(children):
      for child_path, child_leaf in children:
        full_child_path = param_path + child_path
        if isinstance(child_leaf, config.Buildable):
          yield from _make_per_leaf_histories_recursive(child_leaf,
                                                        raw_value_repr,
                                                        full_child_path)
        else:
          yield f'{_path_to_str(full_child_path)} = {child_leaf}'
    else:
      yield _make_per_leaf_history_text(param_path, param_history,
                                        raw_value_repr)

  unset_fields = sorted(
      set(cfg.__signature__.parameters.keys()) -
      set(cfg.__argument_history__.keys()))
  for name in unset_fields:
    if path:
      param_path = tuple(path) + (_ParamName(name),)
    else:
      param_path = (_ParamName(name),)
    yield f'{_path_to_str(param_path)} = <[unset]>'


def _make_per_leaf_history_text(path: _Path,
                                param_history: List[history.HistoryEntry],
                                raw_value_repr: bool) -> str:
  """Returns a string representing a parameter's history.

  Args:
    path: The path to the parameter.
    param_history: The parameter's history.
    raw_value_repr: If True, use `repr` unmodified, otherwise, use a custom
      pretty-printing routine.

  Returns:
    A string representation of the parameter's history that can be displayed to
    the user.
  """
  assert param_history, 'param_history should never be empty.'

  def make_previous_text(entry: history.HistoryEntry) -> str:
    value = _format_value(entry.value, raw_value_repr=raw_value_repr)
    return f'  - previously: {value} @ {entry.location}'

  if len(param_history) > 1:
    past = '\n'.join(map(make_previous_text, reversed(param_history[:-1])))
    past = '\n' + past  # prefix with a newline.
  else:
    past = ''
  current_value = _format_value(
      param_history[-1].value, raw_value_repr=raw_value_repr)
  current = f'{current_value} @ {param_history[-1].location}'
  return f'{_path_to_str(path)} = {current}{past}'



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides utilities for transforming builder functions into `fdl.Config`s.

This module defines the `auto_config` function (and associated helpers), which
can be used to convert an existing function that creates an object graph into a
function that creates a graph of `Config` and `Partial` objects. When the
resulting graph of `Config` and `Partial` objects is built via `fdl.build()`, it
will yield same object graph as the original function.
"""

import ast
import builtins
import functools
import inspect
import textwrap
import types
from typing import Any

from fiddle import config

_CALL_HANDLER_ID = '__auto_config_call_handler__'
_BUILTINS = frozenset([
    builtin for builtin in builtins.__dict__.values()
    if inspect.isroutine(builtin) or inspect.isclass(builtin)
])


def _returns_buildable(signature: inspect.Signature) -> bool:
  """Returns True iff the return annotation is a subclass of config.Buildable."""
  return (signature.return_annotation is not signature.empty and
          inspect.isclass(signature.return_annotation) and
          issubclass(signature.return_annotation, config.Buildable))


def _is_auto_config_eligible(fn_or_cls):
  """Helper to determine if `fn_or_cls` is eligible for auto-config."""
  try:
    signature = inspect.signature(fn_or_cls)
  except ValueError:
    signature = None

  try:
    _ = hash(fn_or_cls)
  except TypeError:
    has_hash = False
  else:
    has_hash = True

  is_buildable = (
      inspect.isclass(fn_or_cls) and issubclass(fn_or_cls, config.Buildable))

  return (
      # We can find a signature...
      signature is not None and
      # It's not a builtin function...
      not inspect.isbuiltin(fn_or_cls) and
      # It's not a builtin type like range()...
      (has_hash and fn_or_cls not in _BUILTINS) and
      # It's not a `fdl.Buildable` already...
      not is_buildable and
      # It's not a method...
      not inspect.ismethod(fn_or_cls) and
      # It's not an `auto_config`ed function...
      not hasattr(fn_or_cls, 'as_buildable') and
      # It's not a function returning a `fdl.Buildable`...
      not _returns_buildable(signature)
  )  # pyformat: disable


def auto_config_call_handler(fn_or_cls, *args, **kwargs):
  """Handles calls in auto_config'ed functions.

  This intercepts calls in an auto-configed function, and determines whether the
  called `fn_or_cls` should be wrapped in a `Config` or `Partial`. If
  `fn_or_cls` is `functools.partial`, the call will instead be converted into a
  call to Fiddle's `Partial`. If it is "auto-config eligible" (see
  `_is_auto_config_eligible`), then a `Config` will be create for `fn_or_cls`
  with the provided arguments. Otherwise, `fn_or_cls` is called directly.

  Args:
    fn_or_cls: The function or class being called.
    *args: The positional arguments with which `fn_or_cls` is being called.
    **kwargs: The keyword arguments with which `fn_or_cls` is being called.

  Returns:
    Depending on `fn_or_cls`, either `Partial`, a `Config`, or the result of
    calling `fn_or_cls` with the provided arguments.
  """
  if fn_or_cls is functools.partial:
    return config.Partial(args[0], *args[1:], **kwargs)

  if _is_auto_config_eligible(fn_or_cls):
    return config.Config(fn_or_cls, *args, **kwargs)

  if hasattr(fn_or_cls, 'as_buildable'):
    return fn_or_cls.as_buildable(*args, **kwargs)

  return fn_or_cls(*args, **kwargs)


class UnsupportedLanguageConstructError(SyntaxError):
  pass


class _AutoConfigNodeTransformer(ast.NodeTransformer):
  """A NodeTransformer that adds the auto-config call handler into an AST."""

  def __init__(self,
               source: str,
               filename: str,
               line_offset: int,
               allow_control_flow=False):
    """Initializes the auto config node transformer instance.

    Args:
      source: The source code of the node that will be transformed by this
        instance. This is used for better error reporting.
      filename: The filename `source` is from.
      line_offset: The line offset of `source` within `filename`.
      allow_control_flow: Whether to permit control flow constructs (loops,
        conditionals, comprehensions, etc). By default, this is `False`, and
        control flow constructs will cause an
        `UnsupportedLanguageConstructError` to be raised.
    """
    self._lines = source.splitlines()
    self._filename = filename
    self._line_offset = line_offset
    self._allow_control_flow = allow_control_flow

    self._function_def_depth = 0

  def _location_for(self, node: ast.AST):
    line_number = node.lineno + self._line_offset
    line = self._lines[node.lineno - 1]
    return (self._filename, line_number, node.col_offset, line)

  def _handle_control_flow(self, node: ast.AST, activatable: bool = False):
    if self._allow_control_flow and activatable:
      return self.generic_visit(node)
    msg = f'Control flow ({type(node).__name__}) is unsupported by auto_config.'
    raise UnsupportedLanguageConstructError(msg, self._location_for(node))

  def _generic_visit_inside_function(self, node):
    try:
      self._function_def_depth += 1
      return self.generic_visit(node)
    finally:
      self._function_def_depth -= 1

  # pylint: disable=invalid-name
  def visit_Call(self, node: ast.Call):
    return ast.Call(
        func=ast.Name(id=_CALL_HANDLER_ID, ctx=ast.Load()),
        args=[node.func, *(self.visit(arg) for arg in node.args)],
        keywords=[self.visit(keyword) for keyword in node.keywords],
    )

  def visit_For(self, node: ast.For):
    return self._handle_control_flow(node, activatable=True)

  def visit_While(self, node: ast.While):
    return self._handle_control_flow(node, activatable=True)

  def visit_If(self, node: ast.If):
    return self._handle_control_flow(node, activatable=True)

  def visit_IfExp(self, node: ast.IfExp):
    return self._handle_control_flow(node, activatable=True)

  def visit_ListComp(self, node: ast.ListComp):
    return self._handle_control_flow(node, activatable=True)

  def visit_SetComp(self, node: ast.SetComp):
    return self._handle_control_flow(node, activatable=True)

  def visit_DictComp(self, node: ast.DictComp):
    return self._handle_control_flow(node, activatable=True)

  def visit_GeneratorExp(self, node: ast.GeneratorExp):
    return self._handle_control_flow(node, activatable=True)

  def visit_Try(self, node: ast.Try):
    return self._handle_control_flow(node)

  def visit_Raise(self, node: ast.Try):
    return self._handle_control_flow(node)

  def visit_With(self, node: ast.With):
    return self._handle_control_flow(node)

  def visit_Yield(self, node: ast.Yield):
    return self._handle_control_flow(node)

  def visit_YieldFrom(self, node: ast.YieldFrom):
    return self._handle_control_flow(node)

  def visit_FunctionDef(self, node: ast.FunctionDef):
    if self._function_def_depth > 0:
      msg = 'Nested function definitions are not supported by auto_config.'
      raise UnsupportedLanguageConstructError(msg, self._location_for(node))
    else:
      return self._generic_visit_inside_function(node)

  def visit_Lambda(self, node: ast.Lambda):
    if self._function_def_depth > 0:
      msg = 'Lambda definitions are not supported by auto_config.'
      raise UnsupportedLanguageConstructError(msg, self._location_for(node))
    else:
      return self._generic_visit_inside_function(node)

  def visit_ClassDef(self, node: ast.ClassDef):
    msg = 'Class definitions are not supported by auto_config.'
    raise UnsupportedLanguageConstructError(msg, self._location_for(node))

  def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
    msg = 'Async function definitions are not supported by auto_config.'
    raise UnsupportedLanguageConstructError(msg, self._location_for(node))

  # pylint: enable=invalid-name


def auto_config(
    fn=None,
    experimental_allow_control_flow=False
) -> Any:  # TODO: More precise return type.
  """Rewrites the given function to make it generate a `Config`.

  This function creates a new function from `fn` by rewriting its AST (abstract
  syntax tree), replacing all `Call` nodes with a custom call handler. When the
  rewritten function is run, the call handler intercepts calls and applies the
  following rules:

    - Calls to builtins, methods, callables without an inferrable signature, or
      other functions that have been `auto_config`ed take place as usual.
    - Calls to `functools.partial` are replaced by calling `fdl.Partial` with
      the same arguments;
    - All other calls are replaced by calling `fdl.Config` with the arguments
      that would have been passed to the called function or class.

  This function may be used standalone or as a decorator. The returned function
  is simply a wrapper around `fn`, but with an additional `as_buildable`
  attribute containing the rewritten function. For example:

      def build_model():
        return Sequential([
            Dense(num_units=128, activation=relu),
            Dense(num_units=128, activation=relu),
            Dense(num_units=1, activation=None),
        ])

      config = auto_config(build_model).as_buildable()

  The resulting `config` is equivalent to the following "manually" constructed
  configuration graph:

      fdl.Config(Sequential, layers=[
          fdl.Config(Dense, num_units=128, activation=relu),
          fdl.Config(Dense, num_units=128, activation=relu),
          fdl.Config(Dense, num_units=1, activation=None),
      ])

  This can then be built with `fdl.build(config)`. Without modification, this
  will result in the same model as just calling `build_model()` directly.
  However, `config` permits changes to the model hyperparameters, for example:

      config.layers[0].num_units = 64
      config.layers[0].activation = 'elu'
      config.layers[1].num_units = 64
      config.layers[1].activation = 'elu'

      modified_model = fdl.build(config)

  Currently, control flow is not supported by default in `auto_config`.
  Experimental support for control flow can be enabled using the
  `experimental_allow_control_flow` argument. If enabled, control flow
  constructs may be used within the function to construct the resulting config
  (for example, a `for` loop could be used to build a list of layers). Control
  flow is never encoded directly as part of the resulting `fdl.Config` (for
  example, there is no `fdl.Config` that will correspond to a conditional or
  loop). While many simple constructs (`for _ in range(10)` etc) work, there
  will also likely be surprising behavior in some circumstances (for example,
  using `itertools` functions in conjunction with a loop will not work, since
  the calls to `itertools` functions will be turned into `fdl.Config` objects).

  Args:
    fn: The function to create a config-generating function from.
    experimental_allow_control_flow: Whether to allow control flow constructs in
      `fn`. By default, control flow constructs will cause an
      `UnsupportedLanguageConstructError` to be thrown.

  Returns:
    A wrapped version of `fn`, but with an additional `as_buildable` attribute
    containing the rewritten function.
  """

  def make_auto_config(fn):
    if not inspect.isfunction(fn):
      raise ValueError('`auto_config` is only compatible with functions.')

    # Get the source code of the function, and remove any indentation which
    # would cause parsing issues when creating the AST (indentation generally is
    # present for nested functions or class methods).
    source = inspect.getsource(fn)
    source = textwrap.dedent(source)

    # Create the NodeTransformer that will transform the AST. The
    # `_AutoConfigNodeTransformer` requires some additional information about
    # the source to provide more informative error messages.
    filename = inspect.getsourcefile(fn)
    line_offset = fn.__code__.co_firstlineno - 1
    node_transformer = _AutoConfigNodeTransformer(
        source=source,
        filename=filename,
        line_offset=line_offset,
        allow_control_flow=experimental_allow_control_flow)

    # Parse the AST, and modify it by intercepting all `Call`s with the
    # `auto_config_call_handler`. Finally, ensure line numbers and code
    # locations match up with the original function, to make errors
    # interpretable.
    node = ast.parse(source)
    node = node_transformer.visit(node)
    node = ast.fix_missing_locations(node)
    node = ast.increment_lineno(node, line_offset)

    # Compile the modified AST, and then find the function code object within
    # the returned module-level code object. Generally, the function is present
    # at `code.co_consts[0]`, but the comprehension below finds it by name.
    # Assuming compilation was successful, the function should really be there,
    # so an assert is used to verify that it was found.
    code = compile(node, inspect.getsourcefile(fn), 'exec')
    code = [
        const for const in code.co_consts  # pytype: disable=attribute-error
        if inspect.iscode(const) and const.co_name == fn.__name__
    ]
    assert len(code) == 1, "Couldn't find modified function code."
    code = code[0]

    # Make sure that the proper globals and closure variables are available to
    # the newly created function, and also add in `auto_config_call_handler`,
    # which is referenced from the modified AST via the `_CALL_HANDLER_ID` (see
    # `ast.Name` node in `_AutoConfigNodeTransformer.visit_Call()`).
    closure_vars = inspect.getclosurevars(fn)
    scope = {
        **fn.__globals__,
        **closure_vars.globals,
        **closure_vars.nonlocals,
        _CALL_HANDLER_ID: auto_config_call_handler,
    }

    # Then, create a function from the compiled function code object, providing
    # the scope.
    auto_config_fn = types.FunctionType(code, scope)
    auto_config_fn.__defaults__ = fn.__defaults__
    auto_config_fn.__kwdefaults__ = fn.__kwdefaults__

    # Finally we wrap the rewritten function to perform additional error
    # checking and enforce that the output is a `fdl.Buildable`.
    @functools.wraps(auto_config_fn)
    def as_buildable(*args, **kwargs):
      output = auto_config_fn(*args, **kwargs)  # pylint: disable=not-callable
      if not isinstance(output, config.Buildable):
        raise TypeError(
            f'The `auto_config` rewritten version of `{fn.__qualname__}` '
            f'returned a `{type(output).__name__}`, which is not a '
            '`fdl.Buildable`. Please ensure this function returns the result '
            'of an `auto_config`-eligible call expression.')
      return output

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
      return fn(*args, **kwargs)

    wrapper.as_buildable = as_buildable

    return wrapper

  # Decorator with empty parenthesis.
  if fn is None:
    return make_auto_config
  else:
    return make_auto_config(fn)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for print."""

import dataclasses
import re
import textwrap
from typing import Union

from absl.testing import absltest
import fiddle as fdl
from fiddle import printing
from fiddle import tagging


class DummyTag(tagging.Tag):
  """Dummy tag for testing."""


class OtherTag(tagging.Tag):
  """Second tag, for fun & profit!"""


def fn_x_y(x, y):  # pylint: disable=unused-argument
  pass


def fn_with_kwargs(x, y, **kwargs):  # pylint: disable=unused-argument
  pass


class TestHelper:

  def __init__(self, a, b):  # pylint: disable=unused-argument
    pass


def fn_with_type_annotations(x: int, y: str, z: float):  # pylint: disable=unused-argument
  pass


@dataclasses.dataclass
class DataclassHelper:
  x: int
  y: str


def advanced_annotations_helper(x: TestHelper, **kwargs: DataclassHelper):  # pylint: disable=unused-argument
  pass


def annotated_kwargs_helper(**kwargs: int):  # pylint: disable=unused-argument
  pass


# Sometimes the module of local symbols shows up differently, depending on how
# this test is imported/run.
_local_module_regex = r'(__main__|fiddle\.printing_test)'


class AsStrFlattenedTests(absltest.TestCase):

  def check_result(self, actual, expected):
    expected = re.escape(textwrap.dedent(expected)).replace(
        '__main__', _local_module_regex)
    self.assertRegex(actual, expected)

  def test_path_to_str(self):
    self.assertEqual(printing._path_to_str(()), '')
    self.assertEqual(
        printing._path_to_str((printing._ParamName('x'), 0)), 'x[0]')
    self.assertEqual(
        printing._path_to_str((printing._ParamName('a'), 2, 'x')), "a[2]['x']")

  def test_simple_printing(self):
    cfg = fdl.Config(fn_x_y, 1, 'abc')
    output = printing.as_str_flattened(cfg)

    self.check_result(output, """\
        x = 1
        y = 'abc'""")

  def test_unset_argument(self):
    cfg = fdl.Config(fn_x_y, 3.14)
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x = 3.14
        y = <[unset]>""")
    self.assertEqual(output, expected)

  def test_nested(self):
    cfg = fdl.Config(fn_x_y, 'x', fdl.Config(fn_x_y, 'nest_x', 123))
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x = 'x'
        y.x = 'nest_x'
        y.y = 123""")
    self.assertEqual(output, expected)

  def test_class(self):
    cfg = fdl.Config(TestHelper, 'a_param', b=123)
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        a = 'a_param'
        b = 123""")
    self.assertEqual(output, expected)

  def test_kwargs(self):
    cfg = fdl.Config(fn_with_kwargs, 1, abc='extra kwarg value')
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x = 1
        y = <[unset]>
        abc = 'extra kwarg value'""")
    self.assertEqual(output, expected)

  def test_nested_kwargs(self):
    cfg = fdl.Config(
        fn_with_kwargs,
        extra=fdl.Config(fn_with_kwargs, 1, nested_extra='whee'))
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x = <[unset]>
        y = <[unset]>
        extra.x = 1
        extra.y = <[unset]>
        extra.nested_extra = 'whee'""")
    self.assertEqual(output, expected)

  def test_nested_collections(self):
    cfg = fdl.Config(fn_x_y,
                     [fdl.Config(fn_x_y, 1, '1'),
                      fdl.Config(TestHelper, 2)])
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x[0].x = 1
        x[0].y = '1'
        x[1].a = 2
        x[1].b = <[unset]>
        y = <[unset]>""")
    self.assertEqual(output, expected)

  def test_multiple_nested_collections(self):
    cfg = fdl.Config(fn_x_y, {
        'a': fdl.Config(fn_with_kwargs, abc=[1, 2, 3]),
        'b': [3, 2, 1]
    }, [fdl.Config(fn_x_y, [fdl.Config(fn_x_y, 1, 2)])])
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x['a'].x = <[unset]>
        x['a'].y = <[unset]>
        x['a'].abc = [1, 2, 3]
        x['b'][0] = 3
        x['b'][1] = 2
        x['b'][2] = 1
        y[0].x[0].x = 1
        y[0].x[0].y = 2
        y[0].y = <[unset]>""")
    self.assertEqual(output, expected)

  def test_default_values(self):

    def test_fn(w, x, y=3, z='abc'):  # pylint: disable=unused-argument
      pass

    cfg = fdl.Config(test_fn, 1)
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        w = 1
        x = <[unset]>
        y = <[unset; default: 3]>
        z = <[unset; default: 'abc']>""")
    self.assertEqual(output, expected)

  def test_tagged_values(self):
    cfg = fdl.Config(fn_x_y, x=DummyTag.new(), y=DummyTag.new(default='abc'))
    output = printing.as_str_flattened(cfg)

    self.check_result(
        output, """\
        x = <[unset]> #__main__.DummyTag
        y = 'abc' #__main__.DummyTag""")

    tagging.set_tagged(cfg, tag=DummyTag, value='cba')
    output = printing.as_str_flattened(cfg)

    self.check_result(
        output, """\
        x = 'cba' #__main__.DummyTag
        y = 'cba' #__main__.DummyTag""")

  def test_tagged_values_multiple_tags(self):
    cfg = fdl.Config(
        fn_x_y,
        x=tagging.TaggedValue(tags=(DummyTag, OtherTag)),
        y=tagging.TaggedValue(tags=(DummyTag, OtherTag), default='abc'))
    output = printing.as_str_flattened(cfg)

    self.check_result(
        output, """\
        x = <[unset]> #__main__.DummyTag #__main__.OtherTag
        y = 'abc' #__main__.DummyTag #__main__.OtherTag""")

    tagging.set_tagged(cfg, tag=DummyTag, value='cba')
    output = printing.as_str_flattened(cfg)

    self.check_result(
        output, """\
        x = 'cba' #__main__.DummyTag #__main__.OtherTag
        y = 'cba' #__main__.DummyTag #__main__.OtherTag""")

  def test_partial(self):
    partial = fdl.Partial(fn_x_y)
    partial.x = 'abc'
    output = printing.as_str_flattened(partial)

    expected = textwrap.dedent("""\
        x = 'abc'
        y = <[unset]>""")
    self.assertEqual(output, expected)

  def test_builtin_types_annotations(self):
    cfg = fdl.Config(fn_with_type_annotations, 1)
    cfg.y = 'abc'
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x: int = 1
        y: str = 'abc'
        z: float = <[unset]>""")
    self.assertEqual(output, expected)

  def test_advanced_type_annotations(self):
    cfg = fdl.Config(advanced_annotations_helper)
    cfg.abc = fdl.Config(DataclassHelper)
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x: TestHelper = <[unset]>
        abc.x: int = <[unset]>
        abc.y: str = <[unset]>""")
    self.assertEqual(output, expected)

  def test_annotated_kwargs(self):
    cfg = fdl.Config(annotated_kwargs_helper, x=1, y='oops')
    output = printing.as_str_flattened(cfg)

    expected = textwrap.dedent("""\
        x: int = 1
        y: int = 'oops'""")
    self.assertEqual(output, expected)

  def test_disabling_type_annotations(self):
    cfg = fdl.Config(fn_with_type_annotations, 1)
    cfg.y = 'abc'
    output = printing.as_str_flattened(cfg, include_types=False)

    expected = textwrap.dedent("""\
        x = 1
        y = 'abc'
        z = <[unset]>""")
    self.assertEqual(output, expected)

  def test_can_print_union_type(self):

    def to_integer(x: Union[int, str]):
      return int(x)

    cfg = fdl.Config(to_integer, 1)
    output = printing.as_str_flattened(cfg, include_types=True)

    expected = textwrap.dedent("""\
        x: typing.Union[int, str] = 1""")
    self.assertEqual(output, expected)

  def test_materialized_default_values(self):

    def test_fn(w, x, y=3, z='abc'):
      del w, x, y, z  # Unused.

    cfg = fdl.Config(test_fn, 1)
    fdl.materialize_defaults(cfg)
    output = printing.as_str_flattened(cfg)
    expected = textwrap.dedent("""\
        w = 1
        x = <[unset]>
        y = 3
        z = 'abc'""")
    self.assertEqual(output, expected)


class HistoryPerLeafParamTests(absltest.TestCase):

  def test_simple_history(self):
    cfg = fdl.Config(fn_x_y, 1, 'abc')
    cfg.x = 2
    output = printing.history_per_leaf_parameter(cfg)
    expected = textwrap.dedent(r"""
        x = 2 @ .*/printing_test.py:\d+:test_simple_history
          - previously: 1 @ .*/printing_test.py:\d+:test_simple_history
        y = 'abc' @ .*/printing_test.py:\d+:test_simple_history""".strip('\n'))
    self.assertRegex(output, expected)

  def test_nested_in_collections(self):
    cfg = fdl.Config(fn_x_y,
                     [fdl.Config(fn_x_y, 1, '1'),
                      fdl.Config(TestHelper, 2)])
    cfg.x[0].x = 3
    cfg.x[1].a = 2  # Reset to same value.
    cfg.x[0].y = 'abc'
    cfg.x[0].x = 4
    output = printing.history_per_leaf_parameter(cfg)
    expected = textwrap.dedent(rf"""
        __fn_or_cls__ = .*fn_x_y .+/printing_test.py:\d+:test_nested_in_collections
        x\[0\].__fn_or_cls__ = .*fn_x_y .+/printing_test.py:\d+:test_nested_in_collections
        x\[0\].x = 4 @ .*/printing_test.py:\d+:test_nested_in_collections
          - previously: 3 @ .*/printing_test.py:\d+:test_nested_in_collections
          - previously: 1 @ .*/printing_test.py:\d+:test_nested_in_collections
        x\[0\].y = 'abc' @ .*/printing_test.py:\d+:test_nested_in_collections
          - previously: '1' @ .*/printing_test.py:\d+:test_nested_in_collections
        x\[1\].__fn_or_cls__ = .*{_local_module_regex}.TestHelper.*/printing_test.py:\d+:test_nested_in_collections
        x\[1\].a = 2 @ .*/printing_test.py:\d+:test_nested_in_collections
          - previously: 2 @ .*/printing_test.py:\d+:test_nested_in_collections
        x\[1\].b = <\[unset\]>
        y = <\[unset\]>""".strip('\n'))
    self.assertRegex(output, expected)

  def test_update_callable_history(self):
    cfg = fdl.Config(fn_x_y, x=1, y=2)
    fdl.update_callable(cfg, fn_with_kwargs)
    cfg.abc = '123'
    output = printing.history_per_leaf_parameter(cfg)
    expected = textwrap.dedent(r"""
        __fn_or_cls__ = .*fn_with_kwargs .+/printing_test.py:\d+:test_update_callable_history
          - previously: .*fn_x_y .+/printing_test.py:\d+:test_update_callable_history
        x = 1 @ .+/printing_test.py:\d+:test_update_callable_history
        y = 2 @ .+/printing_test.py:\d+:test_update_callable_history
        abc = '123' @ .+/printing_test.py:\d+:test_update_callable_history
        kwargs = <\[unset\]>""".strip('\n'))
    self.assertRegex(output, expected)

  def test_materialize_defaults_history(self):

    def test_defaulting_helper(w, x, y=1, z=2):
      del w, x, y, z  # Unused.

    cfg = fdl.Config(test_defaulting_helper, w=0)
    cfg.y = 5
    cfg.y = 6
    fdl.materialize_defaults(cfg)
    output = printing.history_per_leaf_parameter(cfg)
    expected = textwrap.dedent(r"""
        __fn_or_cls__ = .*test_defaulting_helper .+/printing_test.py:\d+:test_materialize_defaults_history
        w = 0 @ .*/printing_test.py:\d+:test_materialize_defaults_history
        y = 6 @ .*/printing_test.py:\d+:test_materialize_defaults_history
          - previously: 5 .*
        z = 2 @ .*/printing_test.py:\d+:test_materialize_defaults_history
        x = <\[unset\]>""".strip('\n'))
    self.assertRegex(output, expected)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for codegen."""
import dataclasses
import functools
from typing import List

from absl.testing import absltest
import fiddle as fdl
from fiddle.codegen import codegen
from fiddle.codegen import test_util
from fiddle.codegen.test_submodule import test_util as submodule_test_util


def tokens(code: str) -> List[str]:
  return code.strip().split()


@dataclasses.dataclass
class Foo:
  a: int
  leaves: List["Foo"] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Bar:
  foo1: Foo
  foo2: Foo


@dataclasses.dataclass
class Baz:
  foo: Foo
  bar: Bar


def identity(x):
  return x


def simple_tree() -> fdl.Config[Foo]:
  return fdl.Config(
      Foo, a=1, leaves=[
          fdl.Config(Foo, a=2),
          fdl.Config(Foo, a=3),
      ])


def shared_config() -> fdl.Config[Baz]:
  shared_foo = fdl.Config(Foo, a=1)

  baz = fdl.Config(Baz)
  baz.foo = shared_foo

  baz.bar = fdl.Config(Bar)
  baz.bar.foo1 = fdl.Config(Foo, a=2)
  baz.bar.foo2 = shared_foo

  return baz


def multi_shared_config() -> fdl.Config[Foo]:
  foo = functools.partial(fdl.Config, Foo)

  shared1 = foo(1)
  shared2 = foo(2, leaves=[shared1])

  return foo(
      3, leaves=[
          shared2,
          foo(4),
          foo(5, leaves=[shared1, shared2]),
      ])


def unshared_child_of_shared() -> fdl.Config[Foo]:
  foo = functools.partial(fdl.Config, Foo)

  shared = foo(1, leaves=[foo(2)])
  return foo(3, leaves=[foo(4, leaves=[shared]), foo(5, leaves=[shared])])


def partial_config() -> fdl.Partial[Baz]:
  # Not sure what it means to share a partial, but let's make sure the code
  # works.
  x = fdl.Partial(Foo, a=1)
  return fdl.Partial(Baz, foo=x, bar=fdl.Partial(Bar, foo1=x))


# Depending on how the test harness invokes this test, the expected values may
# or may not include a module name.
if __name__ == "__main__":
  this_module_import = ""
  this_module_prefix = ""
else:
  this_module_import = "from fiddle.codegen import codegen_test"
  this_module_prefix = "codegen_test."


class CodegenTest(absltest.TestCase):

  def test_map_buildables_plain_tree_preorder(self):
    cfg = simple_tree()
    foos = []

    def pre_traverse_map_fn(buildable: fdl.Buildable):
      foos.append(buildable.a)
      return buildable

    result = codegen.map_buildables(
        buildable=cfg, pre_traverse_fn=pre_traverse_map_fn)
    self.assertEqual(foos, [1, 2, 3])
    self.assertIs(result, cfg)

  def test_map_buildables_plain_tree_postorder(self):
    cfg = simple_tree()
    foos = []

    def post_traverse_map_fn(buildable: fdl.Buildable, new_args):
      del new_args  # unused
      foos.append(buildable.a)
      return buildable

    result = codegen.map_buildables(
        buildable=cfg, post_traverse_fn=post_traverse_map_fn)
    self.assertEqual(foos, [2, 3, 1])
    self.assertIs(result, cfg)

  def test_codegen_dot_syntax_shared(self):
    cfg = shared_config()
    result = codegen.codegen_dot_syntax(cfg)
    expected = f"""
import fiddle as fdl
{this_module_import}


def build_config():
  shared_foo = fdl.Config({this_module_prefix}Foo)
  shared_foo.a = 1

  root = fdl.Config({this_module_prefix}Baz)
  root.foo = shared_foo

  root.bar = fdl.Config({this_module_prefix}Bar)
  root.bar.foo2 = shared_foo

  root.bar.foo1 = fdl.Config({this_module_prefix}Foo)
  root.bar.foo1.a = 2

  return root
    """
    actual_tokens = tokens("\n".join(result.lines()))
    self.assertSequenceEqual(actual_tokens, tokens(expected),
                             "\n".join(result.lines()))

  def test_codegen_multi_shared(self):
    cfg = multi_shared_config()
    result = codegen.codegen_dot_syntax(cfg)
    expected = f"""
import fiddle as fdl
{this_module_import}


def build_config():
  shared_foo = fdl.Config({this_module_prefix}Foo)
  shared_foo.a = 1

  shared_foo_2 = fdl.Config({this_module_prefix}Foo)
  shared_foo_2.a = 2
  shared_foo_2.leaves = [shared_foo]

  root = fdl.Config({this_module_prefix}Foo)
  root.a = 3
  root.leaves = [shared_foo_2, NotImplemented, NotImplemented]  # fdl.Config sub-nodes will replace NotImplemented

  root.leaves[1] = fdl.Config({this_module_prefix}Foo)
  root.leaves[1].a = 4

  root.leaves[2] = fdl.Config({this_module_prefix}Foo)
  root.leaves[2].a = 5
  root.leaves[2].leaves = [shared_foo, shared_foo_2]

  return root
    """
    code = "\n".join(result.lines())
    self.assertSequenceEqual(tokens(code), tokens(expected), code)

  def test_codegen_import_and_exec(self):
    cfg = fdl.Config(
        test_util.Foo, a=1, leaves=[fdl.Config(test_util.Foo, a=2)])
    result = codegen.codegen_dot_syntax(cfg)
    expected = """
import fiddle as fdl
from fiddle.codegen import test_util


def build_config():
  root = fdl.Config(test_util.Foo)
  root.a = 1
  root.leaves = [NotImplemented]  # fdl.Config sub-nodes will replace NotImplemented

  root.leaves[0] = fdl.Config(test_util.Foo)
  root.leaves[0].a = 2

  return root
    """
    code = "\n".join(result.lines())
    self.assertSequenceEqual(tokens(code), tokens(expected), code)

    # Actually run the new builder.
    exec_globals = {}
    exec(code, exec_globals)  # pylint: disable=exec-used
    config = exec_globals["build_config"]()
    foo = fdl.build(config)

    # For now, check the built config values. We might add equality operators to
    # Buildable in the future.
    self.assertEqual(foo.a, 1)
    self.assertLen(foo.leaves, 1)
    self.assertEqual(foo.leaves[0].a, 2)
    self.assertEmpty(foo.leaves[0].leaves)

  def test_codegen_unshared_child_of_shared(self):
    cfg = unshared_child_of_shared()
    result = codegen.codegen_dot_syntax(cfg)
    expected = f"""
import fiddle as fdl
{this_module_import}


def build_config():
  shared_foo = fdl.Config({this_module_prefix}Foo)
  shared_foo.a = 2

  shared_foo_2 = fdl.Config({this_module_prefix}Foo)
  shared_foo_2.a = 1
  shared_foo_2.leaves = [shared_foo]

  root = fdl.Config({this_module_prefix}Foo)
  root.a = 3
  root.leaves = [NotImplemented, NotImplemented]  # fdl.Config sub-nodes will replace NotImplemented

  root.leaves[0] = fdl.Config({this_module_prefix}Foo)
  root.leaves[0].a = 4
  root.leaves[0].leaves = [shared_foo_2]

  root.leaves[1] = fdl.Config({this_module_prefix}Foo)
  root.leaves[1].a = 5
  root.leaves[1].leaves = [shared_foo_2]

  return root
    """
    actual_tokens = tokens("\n".join(result.lines()))
    self.assertSequenceEqual(actual_tokens, tokens(expected),
                             "\n".join(result.lines()))

  def test_codegen_partial(self):
    cfg = partial_config()
    result = codegen.codegen_dot_syntax(cfg)
    expected = f"""
import fiddle as fdl
{this_module_import}


def build_config():
  shared_foo = fdl.Partial({this_module_prefix}Foo)
  shared_foo.a = 1

  root = fdl.Partial({this_module_prefix}Baz)
  root.foo = shared_foo

  root.bar = fdl.Partial({this_module_prefix}Bar)
  root.bar.foo1 = shared_foo

  return root
    """
    actual_tokens = tokens("\n".join(result.lines()))
    self.assertSequenceEqual(actual_tokens, tokens(expected),
                             "\n".join(result.lines()))

  def test_codegen_inner_class_name(self):
    cfg = fdl.Config(test_util.NestedParent.Inner, a=4)
    code = "\n".join(codegen.codegen_dot_syntax(cfg).lines())
    expected = """
import fiddle as fdl
from fiddle.codegen import test_util

def build_config():
  root = fdl.Config(test_util.NestedParent.Inner)
  root.a = 4

  return root
    """
    self.assertSequenceEqual(tokens(code), tokens(expected), code)

  def test_dict_value(self):
    cfg = fdl.Config(identity, x={"foo": fdl.Config(Foo, a=1)})
    code = "\n".join(codegen.codegen_dot_syntax(cfg).lines())
    expected = f"""
import fiddle as fdl
{this_module_import}


def build_config():
  root = fdl.Config({this_module_prefix}identity)
  root.x = {{'foo': NotImplemented}}  # fdl.Config sub-nodes will replace NotImplemented

  root.x['foo'] = fdl.Config({this_module_prefix}Foo)
  root.x['foo'].a = 1

  return root
    """
    self.assertSequenceEqual(tokens(code), tokens(expected), code)

  def test_deeply_nested_constant(self):
    cfg = fdl.Config(identity, x={"bar": [3, 4], "foo": [1, 2]})
    code = "\n".join(codegen.codegen_dot_syntax(cfg).lines())
    expected = f"""
import fiddle as fdl
{this_module_import}


def build_config():
  root = fdl.Config({this_module_prefix}identity)
  root.x = {{'bar': [3, 4], 'foo': [1, 2]}}

  return root
    """
    self.assertSequenceEqual(tokens(code), tokens(expected), code)

  def test_codegen_submodule(self):
    cfg = fdl.Config(
        submodule_test_util.Foo, a=1, leaves=[fdl.Config(test_util.Foo, a=4)])
    code = "\n".join(codegen.codegen_dot_syntax(cfg).lines())
    expected = """
import fiddle as fdl
from fiddle.codegen.test_submodule import test_util
from fiddle.codegen import test_util as test_util_2


def build_config():
  root = fdl.Config(test_util.Foo)
  root.a = 1
  root.leaves = [NotImplemented]  # fdl.Config sub-nodes will replace NotImplemented

  root.leaves[0] = fdl.Config(test_util_2.Foo)
  root.leaves[0].a = 4

  return root
    """
    self.assertSequenceEqual(tokens(code), tokens(expected), code)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fiddle support for types using Lingvo's Params types.

Lingvo's Params types use a somewhat similar pattern to Fiddle Buildables. The
lingvo_params_integration allows you to completely mix-and-match Lingvo types
with arbitrary other Fiddle types. Instead of instantiating a fdl.Config,
instead instantiate a fdl.lingvo.LingvoConfig. You can use the LingvoConfig
identically to a fdl.Config type, including passing it to fdl.build to create
an instance of the type.

Example:

```py

import fiddle as fdl

class ClassA(BaseClass):

  @classmethod
  def Params(cls):  # pylint: disable=invalid-name
    p = super().Params()
    p.Define('a', -1, 'The "a" thing!')
    return p

  def __init__(self, params):
    super().__init__(params)
    self.a = params.a

def main():
  cfg = fdl.lingvo.LingvoConfig(ClassA)
  cfg.a = 42

  obj: ClassA = fdl.build(cfg)
  print(obj.a)  # prints 42
```

You can nest fdl.Config's and fdl.lingvo.LingvoConfig's arbitrarily.
"""

import copy
import inspect
import typing
from typing import Any, Union

from fiddle import config
from lingvo.core import hyperparams
import typing_extensions


class ParamInitable(typing_extensions.Protocol):
  """A protocol for Param-initializable classes."""

  @classmethod
  def Params(cls) -> hyperparams.Params:
    pass

  def __init__(self, params: hyperparams.Params):
    pass


class LingvoParamsAdapter:
  """Adapts a hyperparams.Params instance for use with fdl.Config.

  fdl.Config works with callables, and inspect.Signature will look for a
  `__signature__` property when reflecting on an object. This type takes an
  hyperparams.Params instance and builds an `inspect.Signature` to be stored at
  `self.__signature__`. This wrapped object can then be used with most of the
  `fdl.Config` type entirely unmodified. The wrapped object's `__call__` method
  sets a copy of the Params object with the user's set values and then returns
  the result of the `.Instantiate()` call, mimicking the normal constructor
  behavior from a regular type.

  Attributes:
    params: The wrapped Params instance.
    is_nested: `False` iff the Params instance is the root of a Lingvo config
      tree. (There may be multiple Lingvo config trees within a larger Fiddle
      configuration.)
    __signature__: The inferred signature of a call loosely corresponding to the
      parameters defined in `hyperparams.Params`.
  """
  params: hyperparams.Params
  is_nested: bool
  __signature__: inspect.Signature

  def __init__(self, cls_or_params: Union[ParamInitable, hyperparams.Params]):
    if isinstance(cls_or_params, hyperparams.Params):
      self.params = cls_or_params
      self.is_nested = True
    else:
      self.params = cls_or_params.Params()
      self.is_nested = False
    self.__signature__ = _make_signature_from_lingvo_params(self.params)

  def __call__(self, *args, **kwargs):
    if args:
      raise AssertionError(
          f"*args not empty, please file a bug with repro steps; args: {args}.")
    param_copy = self.params.Copy()  # Always make a copy!
    for arg in kwargs:
      setattr(param_copy, arg, kwargs[arg])
    if self.is_nested:
      return param_copy
    return param_copy.Instantiate()

  def __repr__(self):
    if hasattr(self.params, "cls"):
      return f"<LingvoParamsAdapter({self.params.cls.__qualname__})>"
    return "<LingvoParamsAdapter(<unknown class>)"


class LingvoConfig(config.Config):
  """A Config instance specialized for use with LingvoParam-based types.

  For additional context, see the docstring associated with
  `LingvoParamsAdapter`.
  """

  def __init__(self, params_or_cls: Union["LingvoConfig", LingvoParamsAdapter,
                                          ParamInitable, hyperparams.Params],
               *args, **kwargs):
    if isinstance(params_or_cls, LingvoParamsAdapter):
      # config.Buildable's __deepcopy__ calls `type(self)(self.__fn_or_cls__)`.
      params_or_cls = copy.deepcopy(params_or_cls)
    elif isinstance(params_or_cls, LingvoConfig):
      # config.Buildable's __copy__ calls `type(self)(self)`, and
      # `super().__init__` does the right thing.
      pass
    else:
      params_or_cls = LingvoParamsAdapter(params_or_cls)
    super().__init__(params_or_cls, *args, **kwargs)
    # Must set all param values and not leave them as defaults.
    params = typing.cast(LingvoParamsAdapter, self.__fn_or_cls__).params
    for param_name in params.GetKeys():
      if param_name not in self.__arguments__:
        setattr(self, param_name, params.Get(param_name))

  def __setattr__(self, name: str, value: Any):
    if isinstance(value, hyperparams.Params):
      value = LingvoConfig(value)
    # TODO: what about params hidden inside containers?
    super().__setattr__(name, value)


def _convert_lingvo_param(
    name: str,
    param: hyperparams._Param,  # pylint: disable=protected-access
) -> inspect.Parameter:
  """Returns an inspect.Parameter that corresponds to `param`."""
  return inspect.Parameter(
      name=name,
      kind=inspect.Parameter.KEYWORD_ONLY,
      default=param.GetDefault())


def _make_signature_from_lingvo_params(
    params: hyperparams.Params) -> inspect.Signature:
  """Makes a signature from a lingvo Params instance."""
  parameters = [
      _convert_lingvo_param(name, param)
      for name, param in params._params.items()  # pylint: disable=protected-access
  ]
  return inspect.Signature(parameters)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for autobuilders."""

import dataclasses

from absl.testing import absltest
import fiddle as fdl

from fiddle.experimental.autobuilders import autobuilders as ab
from fiddle.experimental.autobuilders import autobuilders_test_util


@dataclasses.dataclass(frozen=True)
class FakeDense:
  in_dim: int
  out_dim: int


@dataclasses.dataclass(frozen=True)
class FakeDenseSubclass(FakeDense):
  pass


@dataclasses.dataclass(frozen=True)
class FakeMlp:
  first_dense: FakeDense
  second_dense: FakeDense


class RegistryTest(absltest.TestCase):

  def setUp(self):
    autobuilders_test_util.clear_registry()
    super().setUp()

  def test_skeleton_and_get_config(self):

    class Foo:

      def __init__(self, x):
        self.x = x

    @ab.skeleton(Foo)
    def foo_skeleton(config: fdl.Config):  # pylint: disable=unused-variable
      # Note: Setting constants is generally not the purpose of skeletons; we're
      # just doing that here for testing purposes.
      config.x = 3

    foo_builder = ab.config(Foo)
    self.assertIsInstance(foo_builder, fdl.Config)
    foo = fdl.build(foo_builder)
    self.assertIsInstance(foo, Foo)
    self.assertEqual(foo.x, 3)

  def test_get_config_not_present(self):

    class Foo:
      pass

    with self.assertRaisesRegex(KeyError, r".*\bFoo\b"):
      ab.Registry().config(Foo)
    with self.assertRaisesRegex(KeyError, r".*\bFoo\b"):
      ab.config(Foo)

    # Still raises informative error even when table entry is present
    # (for a validator).
    ab.validator(Foo)(lambda config: None)
    with self.assertRaisesRegex(KeyError, r".*\bFoo\b"):
      ab.config(Foo)

  def test_skeleton_registers(self):
    registry = ab.Registry()
    fn = lambda config: None
    registry.skeleton(FakeDense)(fn)
    self.assertDictEqual(registry.table, {
        FakeDense: ab.TableEntry(skeleton_fn=fn, validators=[]),
    })

  def test_skeleton_duplicate_class_registration_error(self):
    registry = ab.Registry()
    registry.skeleton(FakeDense)(lambda config: None)
    with self.assertRaisesRegex(ab.DuplicateSkeletonError, r".*FakeDense.*"):
      registry.skeleton(FakeDense)(lambda config: None)

  def test_skeleton_duplicate_function_registration_error(self):
    # Because the fancy error message includes source lines, make sure that
    # logic doesn't crash with functions too.
    def fake_fn():
      pass

    registry = ab.Registry()
    registry.skeleton(fake_fn)(lambda config: None)
    with self.assertRaisesRegex(ab.DuplicateSkeletonError, r".*fake_fn.*"):
      registry.skeleton(fake_fn)(lambda config: None)

  def test_recursive_skeleton(self):

    @ab.skeleton(FakeDense)
    def dense_skeleton(config: fdl.Config) -> None:  # pylint: disable=unused-variable
      config.in_dim = 4
      config.out_dim = 4

    @ab.skeleton(FakeMlp)
    def mlp_skeleton(config: fdl.Config) -> None:  # pylint: disable=unused-variable
      config.first_dense = ab.config(FakeDense)
      config.first_dense.in_dim = 5
      config.second_dense = ab.config(FakeDense)

    fake_mlp = fdl.build(ab.config(FakeMlp))
    self.assertEqual(
        fake_mlp,
        FakeMlp(
            first_dense=FakeDense(in_dim=5, out_dim=4),
            second_dense=FakeDense(in_dim=4, out_dim=4),
        ))

  def test_auto_skeleton_basic(self):
    ab.auto_skeleton(FakeMlp)
    self.assertEqual(
        ab.config(FakeMlp),
        fdl.Config(FakeMlp, fdl.Config(FakeDense), fdl.Config(FakeDense)))

  def test_auto_skeleton_subclasses_and_existing_skeletons(self):

    @ab.skeleton(FakeDense)
    def dense_skeleton(config: fdl.Config) -> None:  # pylint: disable=unused-variable
      config.in_dim = 4
      config.out_dim = 4

    @ab.skeleton(FakeDenseSubclass)
    def dense_subclass_skeleton(config: fdl.Config) -> None:  # pylint: disable=unused-variable
      config.in_dim = 7
      config.out_dim = 9

    ab.auto_skeleton(FakeMlp, second_dense=FakeDenseSubclass)
    fake_mlp = fdl.build(ab.config(FakeMlp))
    self.assertEqual(
        fake_mlp,
        FakeMlp(
            first_dense=FakeDense(in_dim=4, out_dim=4),
            second_dense=FakeDenseSubclass(in_dim=7, out_dim=9),
        ))

  def test_auto_skeleton_allows_none_as_unconfigured_sentinel(self):
    ab.auto_skeleton(FakeMlp, second_dense=None)
    self.assertEqual(
        ab.config(FakeMlp),
        fdl.Config(
            FakeMlp,
            first_dense=fdl.Config(FakeDense),
        ))

  def test_auto_skeleton_unannotated_raises(self):

    def unannotated(x, y):
      return x + y

    with self.assertRaises(ValueError):
      ab.auto_skeleton(unannotated)

    # But it's fine if we specify defaults, or None
    ab.auto_skeleton(unannotated, x=FakeDense, y=None)

  def test_auto_skeleton_doesnt_configure_primitives(self):

    @dataclasses.dataclass
    class Foo:
      x: int = 4

    with self.assertRaisesRegex(TypeError,
                                r"Parameter 'x'.*cannot be configured"):
      ab.auto_skeleton(Foo)

    # Fine to do by explicitly not configuring the parameter.
    ab.auto_skeleton(Foo, x=None)
    foo = fdl.build(ab.config(Foo))
    self.assertEqual(foo.x, 4)

  def test_auto_skeleton_attribute_skeletons_required(self):
    ab.auto_skeleton(FakeMlp, attributes_require_skeletons=True)

    with self.assertRaisesRegex(KeyError, "No skeleton.*FakeDense"):
      ab.config(FakeMlp)

    def bar(z):
      return z + 1

    def foo(x):
      return x

    ab.auto_skeleton(foo, x=bar, attributes_require_skeletons=True)

    with self.assertRaisesRegex(KeyError, "No skeleton.*bar"):
      ab.config(foo)

  def test_validator_registers(self):
    registry = ab.Registry()
    fn = lambda config: None
    registry.validator(FakeDense)(fn)
    self.assertDictEqual(registry.table, {
        FakeDense: ab.TableEntry(skeleton_fn=None, validators=[fn]),
    })


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Specialized code generation for certain types/values.

Several Python values, such as `jnp.bfloat16` (`jax.numpy.bfloat16`), will
render badly if we just try to call repr() on them. This module creates an
extensible API to register better formatting handlers for ML libraries.
"""

import abc
import dataclasses
from typing import Any, Callable, List

import typing_extensions


class ImportManagerApi(typing_extensions.Protocol):
  """Defines an API for a helper class that imports modules."""

  def add_by_name(self, module_name: str) -> str:
    pass


@dataclasses.dataclass(frozen=True)
class ReprString:
  """Helper class whose repr() is just the provided string."""
  repr_value: str

  def __repr__(self):
    return self.repr_value


class Importable(metaclass=abc.ABCMeta):

  @abc.abstractproperty
  def import_modules(self) -> List[str]:
    raise NotImplementedError()

  @abc.abstractmethod
  def repr_string(self, imported_module_names: List[str]) -> ReprString:
    raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class SingleImportable(Importable):
  """An Importable instance, that only needs a single module."""
  import_module: str
  repr_fn: Callable[[str], str]

  @property
  def import_modules(self) -> List[str]:
    return [self.import_module]

  def repr_string(self, imported_module_names: List[str]) -> ReprString:
    imported_module_name, = imported_module_names
    return ReprString(self.repr_fn(imported_module_name))


_EXACT_VALUE_LOOKUP = {}


def register_exact_value(value: Any, resolution: Importable) -> None:
  """Registers an exact-value match."""
  _EXACT_VALUE_LOOKUP[value] = resolution


def transform_py_value(value: Any, import_manager: ImportManagerApi) -> Any:
  """Transforms a Python value for code generation or printing with repr().

  Args:
    value: Python value, typically a leaf node if there are nested data
      structures. Data structures are not automatically transformed by this
      function.
    import_manager: Import manager used to resolve imports.

  Returns:
    Either `value` unchanged, or a ReprString containing code, which when
    evaluated, returns an equivalent object to `value`. (The exact semantics
    of equivalence--by object identity or regular equality--is not guaranteed,
    and mostly a function of what extensions the user has enabled.)
  """
  try:
    importable: Importable = _EXACT_VALUE_LOOKUP[value]
  except (KeyError, TypeError):
    pass
  else:
    imported_names = [
        import_manager.add_by_name(module_name)
        for module_name in importable.import_modules
    ]
    return importable.repr_string(imported_names)

  # Nothing matched, return the value unchanged and hope its repr() string is
  # valid Python.
  return value



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of Fiddle-internal tools to reflect over modules."""

import inspect
import logging
from typing import Any, List


def find_fiddler_like_things(source_module: Any) -> List[str]:
  """Returns a list of names that look like fiddlers.

  A fiddler is a 1-ary function that mutates a config (and returns None).

  Args:
    source_module: A module upon which to probe for potential fiddler-shaped
      attributes.

  Returns:
    Names of attributes that look like potential fiddlers.
  """
  found_names = []
  for name in dir(source_module):
    if name.startswith('__'):
      continue
    try:
      sig = inspect.signature(getattr(source_module, name))
    except Exception:  # pylint: disable=broad-except
      continue  # Ignore.
    required_args = [
        param for param in sig.parameters.values()
        if _is_non_defaulted_positional_args(param)
    ]
    if len(required_args) == 1:
      found_names.append(name)
  return sorted(found_names)


def _is_non_defaulted_positional_args(param: inspect.Parameter) -> bool:
  """Returns True if `param` is a positional argument with no default."""
  return ((param.kind == param.POSITIONAL_OR_KEYWORD or
           param.kind == param.POSITIONAL_ONLY) and
          param.default is param.empty)


def find_base_config_like_things(source_module: Any) -> List[str]:
  """Returns names of attributes of 0-arity functions that might return Configs.

  A base config-producting function is a function that takes no (required)
  arguments, and returns a `fdl.Buildable`.

  Args:
    source_module: A module upon which to search for `base_config`-like
      functions.

  Returns:
    A list of attributes on `source_module` that appear to be base
    config-producing functions.
  """
  available_base_names = []
  for name in dir(source_module):
    if name.startswith('__'):
      continue
    try:
      sig = inspect.signature(getattr(source_module, name))

      def is_required_arg(name: str) -> bool:
        param = sig.parameters[name]  # pylint: disable=cell-var-from-loop
        return ((param.kind == param.POSITIONAL_ONLY or
                 param.kind == param.POSITIONAL_OR_KEYWORD) and
                param.default is param.empty)

      if not any(filter(is_required_arg, sig.parameters)):
        available_base_names.append(name)
    except Exception:  # pylint: disable=broad-except
      logging.debug(
          'Encountered exception while inspecting function called: %s', name)
  return available_base_names



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tagging."""

import copy
from unittest import mock

from absl.testing import absltest
import fiddle as fdl
from fiddle import tagging
from fiddle import tagging_test_module as tst
from fiddle.experimental import selectors


def return_kwargs(**kwargs):
  return kwargs


def get_single_tag(tagged_value: tagging.TaggedValue) -> tagging.TagType:
  """Gets a single tag from a TaggedValue, errors if there are multiple."""
  assert isinstance(tagged_value, tagging.TaggedValue)
  assert len(tagged_value.tags) == 1
  return next(iter(tagged_value.tags))


class TaggingTest(absltest.TestCase):

  def test_tag_str(self):
    self.assertEqual(
        str(tst.ParameterDType), "#fiddle.tagging_test_module.ParameterDType")
    self.assertEqual(
        str(tst.LinearParamDType),
        "#fiddle.tagging_test_module.LinearParamDType")
    self.assertEqual(
        str(tst.MyModel.EncoderDType),
        "#fiddle.tagging_test_module.MyModel.EncoderDType")
    self.assertEqual(
        str(tst.MyModel.DecoderDType),
        "#fiddle.tagging_test_module.MyModel.DecoderDType")

  def test_tag_instantiation(self):
    with self.assertRaisesRegex(
        TypeError, ".*cannot instantiate Fiddle tags.*ParameterDType.*"):
      _ = tst.ParameterDType()
    with self.assertRaisesRegex(
        TypeError,
        "trying to instantiate fiddle.tagging_test_module.MyModel.DecoderDType"
    ):
      _ = tst.MyModel.DecoderDType()

  def test_missing_tag_description(self):
    with self.assertRaisesRegex(TypeError,
                                "You must provide a tag description"):

      class Foo(tagging.Tag):  # pylint: disable=unused-variable
        pass

  def test_tagvalue_fn_exception(self):
    with self.assertRaisesRegex(
        tagging.TaggedValueNotFilledError,
        "Expected.*TaggedValue.*replaced.*one "
        r"with tags \{fiddle.tagging_test_module.LinearParamDType\} was "
        "not set"):
      tagging.tagvalue_fn({tst.LinearParamDType})

  def test_tagvalue_fn_none_value(self):
    # Test that when the value is None, that's OK.
    self.assertIsNone(
        tagging.tagvalue_fn(tags={tst.ParameterDType}, value=None))

  def test_taggedvalue_default_none(self):
    cfg = tagging.TaggedValue(tags={tst.ParameterDType}, default=None)
    self.assertIsNone(fdl.build(cfg))

  def test_one_taggedvalue_unset_in_config(self):
    cfg = fdl.Config(
        return_kwargs,
        foo=tst.ParameterDType.new(default=None),
        bar=tst.ParameterDType.new())
    with self.assertRaisesRegex(
        tagging.TaggedValueNotFilledError,
        "Expected.*TaggedValue.*replaced.*one "
        r"with tags \{fiddle.tagging_test_module.ParameterDType\} was "
        "not set"):
      fdl.build(cfg)

  def test_set_tagged(self):
    cfg = fdl.Config(
        return_kwargs,
        foo=tst.ParameterDType.new(default=None),
        bar=tst.LinearParamDType.new())
    tagging.set_tagged(cfg, tag=tst.LinearParamDType, value=1)
    self.assertDictEqual(fdl.build(cfg), {"foo": None, "bar": 1})

  def test_set_two_taggedvalues(self):
    cfg = fdl.Config(
        return_kwargs,
        foo=tst.ParameterDType.new(default=None),
        bar=tst.LinearParamDType.new())
    tagging.set_tagged(cfg, tag=tst.ParameterDType, value=1)
    tagging.set_tagged(cfg, tag=tst.LinearParamDType, value=2)
    self.assertDictEqual(fdl.build(cfg), {"foo": 1, "bar": 2})

  def test_double_set_tagged(self):
    cfg = fdl.Config(
        return_kwargs,
        foo=tst.ParameterDType.new(default=None),
        bar=tst.LinearParamDType.new())
    tagging.set_tagged(cfg, tag=tst.LinearParamDType, value=1)
    tagging.set_tagged(cfg, tag=tst.LinearParamDType, value=2)
    self.assertDictEqual(fdl.build(cfg), {"foo": None, "bar": 2})

  def test_set_double_tags(self):
    cfg = fdl.Config(
        return_kwargs,
        foo=tagging.TaggedValue(
            tags=(tst.ParameterDType, tst.LinearParamDType), default=None),
        bar=tst.LinearParamDType.new())
    tagging.set_tagged(cfg, tag=tst.LinearParamDType, value=4)
    self.assertEqual(fdl.build(cfg), {"foo": 4, "bar": 4})

  def test_set_tagged_subclasses(self):
    cfg = fdl.Config(
        return_kwargs,
        foo=tst.LinearParamDType.new(),
        bar=tst.ParameterDType.new())
    tagging.set_tagged(cfg, tag=tst.ParameterDType, value=42)
    self.assertEqual(fdl.build(cfg), {"foo": 42, "bar": 42})

  def test_list_tags(self):
    cfg = fdl.Config(
        return_kwargs,
        foo=tagging.TaggedValue(tags=[tst.ParameterDType], default=None),
        bar=tagging.TaggedValue(tags=[tst.LinearParamDType]))
    tags = tagging.list_tags(cfg)
    self.assertEqual(tags, {tst.LinearParamDType, tst.ParameterDType})

  def test_list_tags_multiple_tags(self):
    cfg = fdl.Config(
        return_kwargs,
        foo=tagging.TaggedValue(tags=(tst.ParameterDType, tst.ActivationDType)),
        bar=tst.LinearParamDType.new())
    tags = tagging.list_tags(cfg)
    self.assertEqual(
        tags, {tst.ParameterDType, tst.ActivationDType, tst.LinearParamDType})

  def test_list_tags_superclasses(self):
    cfg = tst.LinearParamDType.new()
    tags = tagging.list_tags(cfg, add_superclasses=True)
    self.assertEqual(tags, {tst.ParameterDType, tst.LinearParamDType})

  def test_set_only_placeholders_in_subtree(self):
    cfg = fdl.Config(
        return_kwargs,
        output_logits=fdl.Config(
            return_kwargs, dtype=tst.ActivationDType.new(default="float32")),
        encoder=fdl.Config(
            return_kwargs, dtype=tst.ActivationDType.new(default="float32")),
    )

    # At first, the defaults are used.
    self.assertDictEqual(
        fdl.build(cfg), {
            "encoder": {
                "dtype": "float32"
            },
            "output_logits": {
                "dtype": "float32"
            },
        })

    # Setting the dtype for the entire tree overrides both.
    tagging.set_tagged(cfg, tag=tst.ActivationDType, value="bfloat16")
    self.assertDictEqual(
        fdl.build(cfg), {
            "encoder": {
                "dtype": "bfloat16"
            },
            "output_logits": {
                "dtype": "bfloat16"
            },
        })

    # Setting the dtype for the output logits just overrides that one.
    tagging.set_tagged(
        cfg.output_logits, tag=tst.ActivationDType, value="float64")
    self.assertDictEqual(
        fdl.build(cfg), {
            "encoder": {
                "dtype": "bfloat16"
            },
            "output_logits": {
                "dtype": "float64"
            },
        })

  def test_deepcopy_and_object_id(self):
    """Tests that deepcopy() is fine, and key IDs are used, not names."""

    def foo(a, b, c, d):
      return a, b, c, d

    cfg = fdl.Config(foo)
    cfg.a = tst.ParameterDType.new()
    cfg.b = tst.ActivationDType.new()
    cfg.c = tst.ParameterDType.new()
    cfg.d = 4
    copied = copy.deepcopy(cfg)
    copied.d = 40

    tagging.set_tagged(cfg, tag=tst.ParameterDType, value=1)
    tagging.set_tagged(cfg, tag=tst.ActivationDType, value=2)

    self.assertEqual(fdl.build(cfg), (1, 2, 1, 4))

    with self.assertRaises(tagging.TaggedValueNotFilledError):
      fdl.build(copied)

    tagging.set_tagged(copied, tag=tst.ParameterDType, value=10)
    tagging.set_tagged(copied, tag=tst.ActivationDType, value=20)
    self.assertEqual(fdl.build(copied), (10, 20, 10, 40))

  def test_deepcopy_linked_value_object(self):

    def foo(a, b):
      return a, b

    shared_value = fdl.Config(foo, 1, 2)
    cfg = fdl.Config(
        foo,
        a=fdl.TaggedValue(tags={tst.LinearParamDType}, default=shared_value),
        b=fdl.TaggedValue(tags={tst.LinearParamDType}, default=shared_value),
    )
    copied = copy.deepcopy(cfg)
    self.assertIsNot(copied.a, cfg.a)
    self.assertIsNot(copied.a.value, cfg.a.value)

    self.assertIs(get_single_tag(copied.a), get_single_tag(cfg.a))
    self.assertIs(get_single_tag(copied.b), get_single_tag(cfg.b))
    self.assertIs(cfg.a.value, cfg.b.value)
    self.assertIs(copied.a.value, copied.b.value)

  def test_shallow_copy_tagged_values(self):
    cfg = tst.ParameterDType.new()
    copied = copy.copy(cfg)

    self.assertIs(get_single_tag(copied), tst.ParameterDType)
    self.assertIs(copied.value, tagging.NO_VALUE)

    with self.assertRaises(tagging.TaggedValueNotFilledError):
      fdl.build(copied)
    tagging.set_tagged(cfg, tag=tst.ParameterDType, value=1)
    self.assertEqual(fdl.build(cfg), 1)

    with self.assertRaises(tagging.TaggedValueNotFilledError):
      fdl.build(copied)
    tagging.set_tagged(copied, tag=tst.ParameterDType, value=2)
    self.assertEqual(fdl.build(cfg), 1)
    self.assertEqual(fdl.build(copied), 2)

  def test_multiple_keys(self):
    cfg = fdl.TaggedValue(
        tags={tst.ParameterDType, tst.MyModel.EncoderDType}, default=4)
    self.assertEqual(fdl.build(cfg), 4)

    # Set the value using the first key.
    tagging.set_tagged(cfg, tag=tst.ParameterDType, value=5)
    self.assertEqual(fdl.build(cfg), 5)

    # Now set it using the second key.
    tagging.set_tagged(cfg, tag=tst.MyModel.EncoderDType, value=6)
    self.assertEqual(fdl.build(cfg), 6)


# TODO: Test set_tagged that leverages tag inheritance.


class TestWithSelectorMock(TaggingTest):

  def setUp(self):
    super().setUp()

    def new_set_impl(cfg, tag, value):
      selectors.select(cfg, tag=tag).set(value=value)

    self.mock_set = mock.patch.object(tagging, "set_tagged", new_set_impl)
    self.mock_set.start()

  def tearDown(self):
    self.mock_set.stop()
    super().tearDown()


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for auto_config."""

import dataclasses
import functools
import sys
from typing import Any

from absl.testing import absltest
from absl.testing import parameterized

from fiddle import building
from fiddle import config
from fiddle.experimental import auto_config
from fiddle.experimental import autobuilders as ab


def basic_fn(arg, kwarg='test'):
  return {'arg': arg, 'kwarg': kwarg}


def fn_with_kwargs(**kwargs):
  return {'kwargs': kwargs}


@dataclasses.dataclass(frozen=True)
class TestClass:
  arg1: Any
  arg2: Any

  def method(self):
    pass


def explicit_config_building_fn(x: int) -> config.Config:
  return config.Config(basic_fn, 5, kwarg=x)


def pass_through(arg):
  return arg


def _line_number():
  return sys._getframe(1).f_lineno


class AutoConfigTest(parameterized.TestCase):

  def test_create_basic_config(self):
    expected_config = config.Config(TestClass, 1, arg2=2)

    @auto_config.auto_config
    def test_class_config():
      return TestClass(1, 2)

    self.assertEqual(expected_config, test_class_config.as_buildable())
    self.assertEqual(TestClass(1, 2), test_class_config())

  def test_create_basic_config_parents(self):
    expected_config = config.Config(TestClass, 1, arg2=2)

    @auto_config.auto_config()  # Note the parenthesis!
    def test_class_config():
      return TestClass(1, 2)

    self.assertEqual(expected_config, test_class_config.as_buildable())
    self.assertEqual(TestClass(1, 2), test_class_config())

  def test_create_basic_partial(self):
    expected_config = config.Partial(basic_fn, 1, kwarg='kwarg')

    @auto_config.auto_config
    def test_fn_config():
      return functools.partial(basic_fn, 1, 'kwarg')

    self.assertEqual(expected_config, test_fn_config.as_buildable())

  def test_create_config_with_args(self):
    expected_config = config.Config(TestClass, 'positional', arg2='default')

    @auto_config.auto_config
    def test_class_config(arg1, arg2='default'):
      return TestClass(arg1, arg2)

    self.assertEqual(expected_config,
                     test_class_config.as_buildable('positional'))
    self.assertEqual(
        TestClass('positional', 'default'), test_class_config('positional'))

  def test_create_config_with_kwonly_args(self):
    expected_config = config.Config(TestClass, 'positional', arg2='default')

    @auto_config.auto_config
    def test_class_config(arg1, *, arg2='default'):
      return TestClass(arg1, arg2)

    self.assertEqual(expected_config,
                     test_class_config.as_buildable('positional'))
    self.assertEqual(
        TestClass('positional', 'default'), test_class_config('positional'))

  def test_calling_auto_config(self):
    expected_config = config.Config(
        basic_fn, 1, kwarg=config.Config(TestClass, 1, 2))

    @auto_config.auto_config
    def test_class_config():
      return TestClass(1, arg2=2)

    @auto_config.auto_config
    def test_fn_config():
      return basic_fn(1, test_class_config())

    self.assertEqual(expected_config, test_fn_config.as_buildable())
    self.assertEqual({
        'arg': 1,
        'kwarg': TestClass(1, arg2=2)
    }, test_fn_config())

  def test_nested_calls(self):
    expected_config = config.Config(
        TestClass, 1, arg2=config.Config(basic_fn, 2, 'kwarg'))

    @auto_config.auto_config
    def test_class_config():
      return TestClass(1, basic_fn(2, 'kwarg'))

    self.assertEqual(expected_config, test_class_config.as_buildable())
    self.assertEqual(
        TestClass(1, {
            'arg': 2,
            'kwarg': 'kwarg'
        }), test_class_config())

  def test_calling_explicit_function(self):
    expected_config = config.Config(
        TestClass, 1, arg2=config.Config(basic_fn, 5, 10))

    @auto_config.auto_config
    def test_nested_call():
      return TestClass(1, explicit_config_building_fn(10))

    self.assertEqual(expected_config, test_nested_call.as_buildable())

  def test_reference_nonlocal(self):
    nonlocal_var = 3
    expected_config = config.Config(basic_fn, 1, kwarg=nonlocal_var)

    @auto_config.auto_config
    def test_fn_config():
      return basic_fn(1, nonlocal_var)

    self.assertEqual(expected_config, test_fn_config.as_buildable())

  def test_calling_builtins(self):
    expected_config = config.Config(TestClass, [0, 1, 2], ['a', 'b'])

    @auto_config.auto_config
    def test_config():
      return TestClass(list(range(3)), list({'a': 0, 'b': 1}.keys()))

    self.assertEqual(expected_config, test_config.as_buildable())

  def test_auto_config_eligibility(self):
    # Some common types that have no signature.
    for builtin_type in (range, dict):
      self.assertFalse(auto_config._is_auto_config_eligible(builtin_type))
    # Some common builtins.
    for builtin in (list, tuple, sum, any, all, iter):
      self.assertFalse(auto_config._is_auto_config_eligible(builtin))
    self.assertFalse(auto_config._is_auto_config_eligible([].append))
    self.assertFalse(auto_config._is_auto_config_eligible({}.keys))
    # A method.
    test_class = TestClass(1, 2)
    self.assertFalse(auto_config._is_auto_config_eligible(test_class.method))
    # Buildable subclasses.
    self.assertFalse(auto_config._is_auto_config_eligible(config.Config))
    self.assertFalse(auto_config._is_auto_config_eligible(config.Partial))
    # Auto-config annotations are not eligible, and this case is tested in
    # `test_calling_auto_config` above.

  def test_autobuilders_in_auto_config(self):
    expected_config = config.Config(
        basic_fn, arg=ab.config(TestClass, require_skeleton=False))

    @auto_config.auto_config
    def autobuilder_using_fn():
      x = ab.config(TestClass, require_skeleton=False)
      return basic_fn(x)

    self.assertEqual(expected_config, autobuilder_using_fn.as_buildable())

  def test_auto_configuring_non_function(self):
    with self.assertRaisesRegex(ValueError, 'only compatible with functions'):
      auto_config.auto_config(3)

  def test_require_buildable_return_type(self):

    @auto_config.auto_config
    def test_config():
      pass

    with self.assertRaisesRegex(
        TypeError, r'The `auto_config` rewritten version of '
        r'`AutoConfigTest\.\w+\.<locals>\.test_config` '
        r'returned a `NoneType`, which is not a `fdl\.Buildable`\. Please '
        r'ensure this function returns the result of an `auto_config`-eligible '
        r'call expression\.'):
      test_config.as_buildable()

  def test_control_flow_if(self):

    def test_config(condition):
      if condition:
        return pass_through('true branch')
      else:
        return pass_through('false branch')

    error_line_number = _line_number() - 5

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Control flow \(If\) is unsupported by auto_config\. '
        r'\(auto_config_test\.py, line \d+\)') as context:
      auto_config.auto_config(test_config)

    self.assertEqual(error_line_number, context.exception.lineno)
    # The text and offset here don't quite match with the indentation level of
    # test_config above, due to the `dedent` performed inside `auto_config`.
    # This slight discrepancy is unlikely to cause any confusion.
    self.assertEqual('  if condition:', context.exception.text)
    self.assertEqual(2, context.exception.offset)

    build_config_fn = auto_config.auto_config(
        test_config, experimental_allow_control_flow=True).as_buildable
    self.assertEqual('true branch', building.build(build_config_fn(True)))
    self.assertEqual('false branch', building.build(build_config_fn(False)))

  def test_control_flow_for(self):

    def test_config():
      layers = []
      for i in range(3):
        layers.append(TestClass(i, i))
      return pass_through(layers)

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Control flow \(For\) is unsupported by auto_config\.'):
      auto_config.auto_config(test_config)

    expected_config = config.Config(
        pass_through, [config.Config(TestClass, i, i) for i in range(3)])
    actual_config = auto_config.auto_config(
        test_config, experimental_allow_control_flow=True).as_buildable()
    self.assertEqual(expected_config, actual_config)

  def test_control_flow_while(self):

    def test_config():
      i = 10
      while i > 0:
        i -= 1
      return pass_through(i)

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Control flow \(While\) is unsupported by auto_config\.'):
      auto_config.auto_config(test_config)

    pass_through_config = auto_config.auto_config(
        test_config, experimental_allow_control_flow=True).as_buildable()
    self.assertEqual(0, building.build(pass_through_config))

  @parameterized.parameters(
      (lambda: pass_through([i + 1 for i in [0, 1, 2]]), 'ListComp'),
      (lambda: pass_through({i + 1 for i in [0, 1, 2]}), 'SetComp'),
      (lambda: pass_through({i + 1: None for i in [0, 1, 2]}), 'DictComp'),
      (lambda: pass_through((i + 1 for i in [0, 1, 2])), 'GeneratorExp'),
      (lambda: pass_through(None if list() else [1, 2, 3]), 'IfExp'),
  )
  def test_control_flow_expression(self, test_config, node_name):
    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        rf'Control flow \({node_name}\) is unsupported by auto_config\.'):
      auto_config.auto_config(test_config)

    pass_through_config = auto_config.auto_config(
        test_config, experimental_allow_control_flow=True).as_buildable()
    self.assertCountEqual([1, 2, 3], building.build(pass_through_config))

  def test_allow_control_flow_decorator(self):

    @auto_config.auto_config(experimental_allow_control_flow=True)
    def test_config():
      return pass_through([i + 1 for i in range(3)])

    self.assertEqual([1, 2, 3], building.build(test_config.as_buildable()))

  def test_disallow_try(self):

    def test_config():
      try:
        pass
      except:  # pylint: disable=bare-except
        pass

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Control flow \(Try\) is unsupported by auto_config\.'):
      auto_config.auto_config(test_config)

  def test_disallow_raise(self):

    def test_config():
      raise Exception()

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Control flow \(Raise\) is unsupported by auto_config\.'):
      auto_config.auto_config(test_config)

  def test_disallow_with(self):

    def test_config():
      with open(''):
        pass

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Control flow \(With\) is unsupported by auto_config\.'):
      auto_config.auto_config(test_config)

  def test_disallow_yield(self):

    def test_config():
      yield

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Control flow \(Yield\) is unsupported by auto_config\.'):
      auto_config.auto_config(test_config)

  def test_disallow_yield_from(self):

    def test_config():
      yield from (i for i in range(5))

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Control flow \(YieldFrom\) is unsupported by auto_config\.'):
      auto_config.auto_config(test_config)

  def test_disallow_function_definitions(self):

    def test_config():

      def nested_fn():
        pass

      return nested_fn()

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Nested function definitions are not supported by auto_config\.'):
      auto_config.auto_config(test_config)

  def test_disallow_class_definitions(self):

    def test_config():

      class NestedClass:
        pass

      return NestedClass()

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Class definitions are not supported by auto_config\.'):
      auto_config.auto_config(test_config)

  def test_disallow_lambda_definitions(self):

    def test_config():
      return pass_through(lambda: TestClass(1, 2))

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Lambda definitions are not supported by auto_config\.'):
      auto_config.auto_config(test_config)

  def test_disallow_async_function_definitions(self):

    async def test_config():
      pass

    with self.assertRaisesRegex(
        auto_config.UnsupportedLanguageConstructError,
        r'Async function definitions are not supported by auto_config\.'):
      auto_config.auto_config(test_config)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for namespace_config."""

import types

from absl.testing import absltest
import fiddle as fdl
from fiddle.experimental import namespace_config


class NamespaceConfigTest(absltest.TestCase):

  def test_simple(self):
    cfg = namespace_config.NamespaceConfig()
    cfg.x = 1
    cfg.y = 'abc'
    obj = fdl.build(cfg)
    self.assertIsInstance(obj, types.SimpleNamespace)
    self.assertEqual(1, obj.x)
    self.assertEqual('abc', obj.y)

  def test_initialize_from_dict(self):
    test_dict = {'a': 1, 'b': 'xyz'}
    cfg = namespace_config.NamespaceConfig(**test_dict)
    obj = fdl.build(cfg)
    self.assertIsInstance(obj, types.SimpleNamespace)
    self.assertEqual(1, obj.a)
    self.assertEqual('xyz', obj.b)

  def test_nesting(self):

    def dummy_fn(**kwargs):
      return (dummy_fn, kwargs)

    cfg = namespace_config.NamespaceConfig()
    cfg.nested = fdl.Config(dummy_fn, x='a')
    cfg.nested.y = 123
    cfg.nested.other_dictconfig = namespace_config.NamespaceConfig(
        very_inside='here')
    cfg.other = 321

    obj = fdl.build(cfg)
    self.assertIsInstance(obj, types.SimpleNamespace)
    self.assertIs(dummy_fn, obj.nested[0])
    self.assertEqual(321, obj.other)
    self.assertIsInstance(obj.nested[1]['other_dictconfig'],
                          types.SimpleNamespace)
    self.assertEqual('here', obj.nested[1]['other_dictconfig'].very_inside)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Routines for nice string representations of cfg values.

Some types (e.g. numpy arrays, jax dtypes, TensorFlow, etc) are often referred
unambiguously by a name that differs from their qualified path. This library
exposes the pretty-printing functionality that powers the codegen features to
other string representations, such as graphviz or string serialization.
"""

from typing import Any

from fiddle.codegen import codegen
from fiddle.codegen import special_value_codegen


class LazyImportManager(special_value_codegen.ImportManagerApi):
  """Import manager for GraphViz & string formatting.

  We create an instance of the codegen's import manager on every method call,
  which effectively will alias imports like 'jax.numpy' to 'jnp', but will not
  create new import aliases when two modules of the same name are referenced.
  """

  def add_by_name(self, module_name: str) -> str:
    """Returns the usable name generated from an ephemeral import manager."""
    namespace = codegen.Namespace()
    import_manager = codegen.ImportManager(namespace)
    return import_manager.add_by_name(module_name)


def pretty_print(value: Any) -> str:
  """Returns the nicest eval-able string representation for `value`."""
  return special_value_codegen.transform_py_value(value, LazyImportManager())



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""History tracking for config objects.

In order to make it easier to understand how config objects have been
constructed, config objects keep track of the sequence of changes made to them.
This file contains the machinery to record this history and easily manipulate
it.

The history functionality has a few bits of non-trivial logic:

 1. Each history entry is associated with a monotonically increasing number.
    This allows sequencing history across multiple config objects. The
    implementation of this capability is guaranteed to be threadsafe, so users
    can mutate (different!) config objects from multiple threads without fear.
 2. The logic to determine where in the user's source code a mutation was
    generated is configurable. Simply use `custom_location` to temporarily set a
    different location provider; see the documentation on the `custom_location`
    context manager for an example.
"""

import contextlib
import dataclasses
import itertools
import traceback
from typing import Any, Callable, Iterator, Optional, Union

# An incrementing counter to allow for time-travel debugging.
_set_counter = itertools.count()


@dataclasses.dataclass(frozen=True)
class Location:
  """Information about where a parameter was set."""
  filename: str
  line_number: int
  function_name: Optional[str]

  def __str__(self) -> str:
    if self.function_name is None:
      return f"{self.filename}:{self.line_number}"
    return f"{self.filename}:{self.line_number}:{self.function_name}"


# A function that returns a location.
LocationProvider = Callable[[], Location]


def _stacktrace_location_provider() -> Location:
  """Returns a string corresponding to the user-function that set the field.

  Raises:
    RuntimeError: if no suitable stack frame can be found.
  """
  for frame in reversed(traceback.extract_stack()):
    if (not frame.filename.endswith("fiddle/config.py") and
        not frame.filename.endswith("fiddle/history.py") and
        not frame.filename.endswith("fiddle/materialize.py")):
      return Location(
          filename=frame.filename,
          line_number=frame.lineno,
          function_name=frame.name)
  raise RuntimeError("Cannot find a suitable frame in the stack trace!")


# The location provider to use when instantiating new HistoryEntry's.
_location_provider: LocationProvider = _stacktrace_location_provider


class _Sentinel:
  """A unique marker class (and instance)."""
  pass


# A marker object to record when a field was deleted.
DELETED = _Sentinel()


@dataclasses.dataclass(frozen=True)
class HistoryEntry:
  """An entry in the history table for a config object.

  Attributes:
    sequence_id: The global sequence number, to allow ordering of history
      sequences across config objects.
    param_name: The parameter name for the history entry.
    value: The new value of the field, or _unset_sentinal if the field was
      `del`d.
    location: The location in user code that made the modification.
  """
  sequence_id: int
  param_name: str
  value: Union[Any, _Sentinel]
  location: Location


def entry(param_name: str, value: Union[_Sentinel, Any]) -> HistoryEntry:
  """Returns a newly constructed history entry.

  Args:
    param_name: The name of the config parameter the entry is associated with.
    value: The new value, or DELETED to indicate a deletion of the param.

  Returns:
    The newly constructed HistoryEntry.
  """
  return HistoryEntry(
      sequence_id=next(_set_counter),
      param_name=param_name,
      value=value,
      location=_location_provider())


@contextlib.contextmanager
def custom_location(
    temporary_provider: LocationProvider) -> Iterator[LocationProvider]:
  """Temporarily sets a custom LocationProvider.

  Example usage:
  ```py
  my_config =  # ...
  with custom_location(lambda: Location('my_loc', 123, None)):
    my_config.x = 123  # the location of my_config.x will be 'my_loc:123'.
  ```

  Args:
    temporary_provider: A location provider to use for the duration of the with
      block.

  Yields:
    The temporary provider.
  """
  global _location_provider
  original_location_provider = _location_provider
  _location_provider = temporary_provider
  yield temporary_provider
  _location_provider = original_location_provider



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library for generating code from a Config or Partial object."""

import collections
import dataclasses
import inspect
import itertools
import keyword
import re
from typing import Any, Callable, Dict, List, Sequence, Set

from absl import logging
from fiddle import config as fdl
from fiddle.codegen import mini_ast
from fiddle.codegen import special_value_codegen
from fiddle.experimental import daglish
import tree


def _camel_to_snake(name: str) -> str:
  """Converts a camel or studly-caps name to a snake_case name."""
  return re.sub(r"(?<=.)([A-Z])", lambda m: "_" + m.group(0).lower(),
                name).lower()


def map_buildables(
    buildable: fdl.Buildable,
    pre_traverse_fn: Callable[[fdl.Buildable], fdl.Buildable] = lambda x: x,
    post_traverse_fn: Callable[[fdl.Buildable, Dict[str, Any]],
                               fdl.Buildable] = lambda x, new_args: x,
    leaf_fn: Callable[[Any], Any] = lambda x: x,
) -> fdl.Buildable:
  """Maps over a tree of builders.

  Args:
    buildable: Config or Partial instance.
    pre_traverse_fn: Function applied before traversing into a child.
    post_traverse_fn: Function applied after traversing into a child.
    leaf_fn: Function applied to transform any leaf values.

  Returns:
    Dictionary of new arguments, after mapping with this function.
  """

  def map_fn(leaf):
    if isinstance(leaf, fdl.Buildable):
      map_buildables(leaf, pre_traverse_fn, post_traverse_fn, leaf_fn)
    else:
      return leaf_fn(leaf)

  buildable = pre_traverse_fn(buildable)
  new_args = tree.map_structure(map_fn, buildable.__arguments__)
  return post_traverse_fn(buildable, new_args)


# Project-specific import aliases.
_SPECIAL_IMPORT_ALIASES = {}


def register_import_alias(name: str, import_stmt: mini_ast.ImportNode) -> None:
  """Registers an import alias.

  Typically this is called by extensions in `fiddle.extensions`.

  Args:
    name: Full module name to alias. Often, this is what can be found in
      `type(py_value).__module__.__name__`.
    import_stmt: Import statement to emit for this module.
  """
  _SPECIAL_IMPORT_ALIASES[name] = import_stmt


def _get_import_aliases() -> Dict[str, mini_ast.ImportNode]:
  """Dictionary of import aliases."""
  return {
      **_SPECIAL_IMPORT_ALIASES,
      "fiddle.config":
          mini_ast.ImportAs(name="fdl", module="fiddle"),
  }


@dataclasses.dataclass
class Namespace:
  """Manages active Python instance names."""

  names: Set[str] = dataclasses.field(
      default_factory=lambda: set(keyword.kwlist))

  def __contains__(self, key: str) -> bool:
    return key in self.names

  def add(self, name: str) -> str:
    """Adds a name, checking if it already exists, and returns it."""
    if name in self.names:
      raise ValueError(
          f"Tried to add {name!r} (e.g. an import), but it already exists!")
    self.names.add(name)
    return name

  def get_new_name(self, base_name, prefix: str = "shared_") -> str:
    """Generates a new name for a given instance.

    Example:
      self.names = ["foo"]
      get_new_name("bar") == "shared_bar"

    Example 2:
      self.names = ["shared_foo", "bar"]
      get_new_name("Foo") == "shared_foo_2"

    Args:
      base_name: Base name to derive a new name for. This will be converted to
        snake case automatically.
      prefix: Prefix to prepend to any names.

    Returns:
      New unique name for a variable representing a config for that function
      or class.
    """

    name = prefix + _camel_to_snake(base_name)
    if name not in self.names:
      return self.add(name)
    for i in itertools.count(start=2):
      if f"{name}_{i}" not in self.names:
        return self.add(f"{name}_{i}")
    raise AssertionError("pytype helper -- itertools.count() is infinite")


@dataclasses.dataclass
class ImportManager:
  """Helper class to maintain a list of import statements."""

  namespace: Namespace
  imports: List[mini_ast.ImportNode] = dataclasses.field(default_factory=list)
  aliases: Dict[str, mini_ast.ImportNode] = dataclasses.field(
      default_factory=_get_import_aliases)

  def add_by_name(self, module_name: str) -> str:
    """Adds an import given a module name.

    This is a slightly lower-level API than `add`; you should only use it if
    you don't have access.

    Args:
      module_name: String module name to try to import.

    Returns:
      Alias for the imported module.
    """
    result = self.aliases.get(module_name)
    if not result:
      if "." in module_name:
        parent, module_name = module_name.rsplit(".", 1)
        result = mini_ast.FromImport(name=module_name, parent=parent)
      else:
        result = mini_ast.DirectImport(module_name)

    if result not in self.imports:
      if result.name in self.namespace:
        # Create or adjust the alias for the import.
        new_name = self.namespace.get_new_name(result.name, prefix="")
        result = result.change_alias(new_name)
      else:
        self.namespace.add(result.name)
      self.imports.append(result)
    return result.name

  def add(self, fn_or_cls: Any) -> str:
    """Adds an import if it doesn't exist.

    This adds an import statement to this manager.

    Args:
      fn_or_cls: Function or class.

    Returns:
      Relative-qualified name for the instance.
    """
    module_name = inspect.getmodule(fn_or_cls).__name__
    fn_or_cls_name = fn_or_cls.__qualname__
    if module_name == "__main__":
      logging.warning(
          "%s's module is __main__, so an import couldn't be added.",
          fn_or_cls_name)
      return fn_or_cls_name

    imported_name = self.add_by_name(module_name)
    return f"{imported_name}.{fn_or_cls_name}"

  def sorted_imports(self):
    """Returns imports sorted lexicographically."""
    return sorted(self.imports, key=lambda x: x.sortkey())


def assignment_path(base_var: str, path: Sequence[daglish.PathElement]) -> str:
  """Generates the LHS of an assignment, given a traversal path.

  Example: ["foo", 3, "bar"] -> "foo[3].bar".

  Args:
    base_var: Base variable name.
    path: Attribute path on `base_var` to assign to.

  Returns:
    Python code string for the LHS of an assignment.

  Raises:
    TypeError: If the first path element is not a string, or if any path element
      is not a string or an int.
  """

  return base_var + "".join(x.code for x in path)


@dataclasses.dataclass(frozen=True)
class _VarReference:
  """Helper class whose repr() is just the provided string."""
  name: str

  def __repr__(self):
    return self.name


def _get_shared_buildables(buildable: fdl.Buildable) -> List[fdl.Buildable]:
  """Finds any sub-buildable nodes which are referenced by multiple parents."""
  # Find shared nodes.
  to_count = collections.Counter()
  children_by_id = {}

  def _count_map_fn(child: fdl.Buildable):
    to_count[id(child)] += 1
    children_by_id[id(child)] = child
    return child

  map_buildables(buildable, _count_map_fn)
  return [
      children_by_id[child_hash]
      for child_hash, count in to_count.items()
      if count > 1
  ]


def _has_child_buildables(value: Any) -> bool:
  result = False

  def leaf_fn(leaf):
    nonlocal result
    result = result or isinstance(leaf, fdl.Buildable)

  tree.map_structure(leaf_fn, value)
  return result


@dataclasses.dataclass
class SharedBuildableManager:
  """Helper class to manage shared configuration objects."""

  namespace: Namespace
  import_manager: ImportManager
  instances: List[mini_ast.CodegenNode] = dataclasses.field(
      default_factory=list)
  instance_names_by_id: Dict[int, str] = dataclasses.field(default_factory=dict)

  def __contains__(self, buildable: fdl.Buildable):
    return id(buildable) in self.instance_names_by_id

  def add(self, name: str, buildable: fdl.Buildable,
          decl: mini_ast.CodegenNode) -> None:
    """Adds a shared instance.

    Args:
      name: Variable name for code representing this instance.
      buildable: Actual Python fdl.Buildable object represented by the code
        being generated.
      decl: Code declaration for this instance.
    """
    self.instances.append(decl)
    self.instance_names_by_id[id(buildable)] = name

  def assign(self, lhs_var: str, lhs_path: Sequence[daglish.PathElement],
             attr_value: Any) -> mini_ast.CodegenNode:
    """Returns an assignment for a Python value.

    When this value references shared builder objects, then those objects are
    replaced with name references to the shared objects.

    Args:
      lhs_var: Variable name of the left-hand side assignment.
      lhs_path: Attribute path on `lhs_var` to assign to.
      attr_value: Python value representing the right-hand side of the
        expression.

    Returns:
      Codegen node representing the assignment.
    """

    if lhs_path and isinstance(lhs_path[-1], daglish.ListOrTupleItem):
      # Skip if we're re-assigning to a list element. The only case when we want
      # to override list elements is when we're assigning to sub-configurations,
      # which were initially assigned NotImplemented. However, these are set in
      # the top stanza of configure_main_tree_block(), not in calls to this
      # method, so we can just return here without further checks.
      return mini_ast.Noop()

    used_not_implemented = False

    def _map_fn(child):
      nonlocal used_not_implemented
      if child in self:
        return _VarReference(self.instance_names_by_id[id(child)])
      elif isinstance(child, fdl.Buildable):
        used_not_implemented = True
        return _VarReference("NotImplemented")
      else:
        return special_value_codegen.transform_py_value(child,
                                                        self.import_manager)

    lhs = assignment_path(lhs_var, lhs_path)
    assignment = mini_ast.Assignment(
        lhs, repr(tree.map_structure(_map_fn, attr_value)))
    if used_not_implemented:
      return mini_ast.TrailingComment(
          assignment, "fdl.Config sub-nodes will replace NotImplemented")
    else:
      return assignment


def _configure_shared_object(
    buildable: fdl.Buildable,
    shared_manager: SharedBuildableManager,
    import_manager: ImportManager,
    variable_name_prefix: str = "shared_",
) -> None:
  """Generates configuration for a shared object.

  The objects configured will be added to `shared_manager`.

  Args:
    buildable: Buildable for which we will generate configuration code. Its
      dependencies can form a DAG, and any leaf nodes will be added first.
    shared_manager: Shared object manager.
    import_manager: Import manager.
    variable_name_prefix: Prefix for any variables introduced.
  """

  def _inner(child: fdl.Buildable, new_args: Dict[str, Any]) -> fdl.Buildable:
    """Generates code for a shared instance."""
    if child in shared_manager:
      # Already added by another pass (for chain dependencies of shared
      # objects).
      return child

    del new_args  # unused
    # Name this better..
    name = shared_manager.namespace.get_new_name(
        child.__fn_or_cls__.__name__, prefix=variable_name_prefix)
    relname = import_manager.add(child.__fn_or_cls__)
    buildable_subclass_str = import_manager.add(child.__class__)
    nodes = [mini_ast.Assignment(name, f"{buildable_subclass_str}({relname})")]
    for key, value in child.__arguments__.items():
      path = [daglish.BuildableAttributeItem(child, key)]
      nodes.append(shared_manager.assign(name, path, value))
    shared_manager.add(name, child, mini_ast.ImmediateAttrsBlock(nodes))
    return child

  map_buildables(buildable, post_traverse_fn=_inner)


def codegen_dot_syntax(buildable: fdl.Buildable) -> mini_ast.CodegenNode:
  """Generates code, preferring nested dot-attribute assignment when possible.

  Example code output (abbreviated and with additional comments):

  # Shared instances are on top, since this format supports DAGs. However
  # choosing names is hard.
  shared_foo = fdl.Config(Foo)
  shared_foo.a = 1

  # Subsequent blocks set up a tree, doing attribute dot-assignment.
  root = fdl.Config(Baz)
  root.foo = shared_foo
  root.bar = fdl.Config(Bar)
  root.bar.foo2 = shared_foo

  Args:
    buildable: Config or Partial instance.

  Returns:
    Codegen node representing output code.
  """
  namespace = Namespace()
  import_manager = ImportManager(namespace)
  shared_manager = SharedBuildableManager(
      namespace, import_manager=import_manager)

  # In this method, we will configure any shared objects. This method is fully
  # DAG compliant, and we'd consider using it for the whole codegen, if it were
  # easier to name things.
  for shared_obj in _get_shared_buildables(buildable):
    _configure_shared_object(
        shared_obj,
        shared_manager=shared_manager,
        import_manager=import_manager,
    )

  # Once we have shared objects detected and set up, the rest of the config
  # becomes a tree.
  main_tree_blocks = []

  def configure_main_tree_block(child: fdl.Buildable, path: List[Any]):
    """Configures a tree node for the main configuration block."""
    relname = import_manager.add(child.__fn_or_cls__)
    buildable_subclass_str = import_manager.add(child.__class__)
    nodes = [
        mini_ast.Assignment(
            assignment_path("root", path),
            f"{buildable_subclass_str}({relname})")
    ]
    deferred = []  # Defer configuring sub-Buildable nodes.

    def handle_child_attr(sub_path, value):
      """Inner handler for traverse_with_path, assigning attributes."""

      if not sub_path:
        # Skip top level __arguments__ dict.
        return

      # Convert sub_path to a daglish path. In the future, this traversal will
      # be part of daglish.
      arg_dict_key, *rest = sub_path
      rest_path_elts = []
      for tree_lib_elt in rest:
        if isinstance(tree_lib_elt, int):
          rest_path_elts.append(daglish.ListOrTupleItem([], tree_lib_elt))
        else:
          rest_path_elts.append(daglish.DictItem({}, tree_lib_elt))
      full_path = [
          *path,
          daglish.BuildableAttributeItem(child, arg_dict_key), *rest_path_elts
      ]

      if rest_path_elts and not _has_child_buildables(value):
        pass
      elif isinstance(value, fdl.Buildable) and value not in shared_manager:
        deferred.append((value, full_path))
      else:
        nodes.append(shared_manager.assign("root", full_path, value))

    tree.traverse_with_path(handle_child_attr, child.__arguments__)
    main_tree_blocks.append(mini_ast.ImmediateAttrsBlock(nodes))

    # Recurses to configure sub-Buildable nodes.
    for sub_child, sub_path in deferred:
      configure_main_tree_block(sub_child, sub_path)

  configure_main_tree_block(buildable, [])

  # Adds the final return statement, glues together the shared instance block
  # with the main tree block, and adds imports.
  main_tree_blocks.append(mini_ast.ReturnStmt("root"))
  return mini_ast.ConfigBuilder(import_manager.sorted_imports(), [
      mini_ast.SharedThenResultAssignment(shared_manager.instances,
                                          main_tree_blocks)
  ])



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Very minimal AST library.

These nodes are not super useful compared to something like libcst, since they
often have opaque string values. Their purpose is primarily to help with the
formatting of Python code.
"""

import abc
import dataclasses
from typing import List, Union


class CodegenNode(metaclass=abc.ABCMeta):
  """Super-lightweight codegen node."""

  @abc.abstractmethod
  def lines(self) -> List[str]:
    """Returns lines of code for this element.

    If the final lines should be indented, any outer nodes should modify the
    output of this method.

    Returns:
      Lines of code.
    """
    pass


@dataclasses.dataclass(frozen=True)
class Noop(CodegenNode):
  """No-op node, when emitting something is convenient."""

  def lines(self) -> List[str]:
    return []


@dataclasses.dataclass(frozen=True)
class ImportNode(CodegenNode):
  name: str  # Name of the module; should precede any class names in code.

  @abc.abstractmethod
  def sortkey(self) -> str:
    """Returns a string module key, which can be used for sorting imports."""
    raise NotImplementedError()

  @abc.abstractmethod
  def change_alias(self, alias_name: str) -> "ImportNode":
    """Creates a version of this import with an alias (`import ... as`)."""


@dataclasses.dataclass(frozen=True)
class DirectImport(ImportNode):
  """Imports a module as a fully-qualified name."""

  def lines(self) -> List[str]:
    return [f"import {self.name}"]

  def sortkey(self) -> str:
    return self.name

  def change_alias(self, alias_name: str) -> ImportNode:
    return ImportAs(alias_name, module=self.name)


@dataclasses.dataclass(frozen=True)
class FromImport(ImportNode):
  """Imports a module name from a base package."""

  parent: str

  def lines(self) -> List[str]:
    return [f"from {self.parent} import {self.name}"]

  def sortkey(self) -> str:
    return f"{self.parent}.{self.name}"

  def change_alias(self, alias_name: str) -> ImportNode:
    return FromImportAs(alias_name, parent=self.parent, module=self.name)


@dataclasses.dataclass(frozen=True)
class ImportAs(ImportNode):
  """Same as DirectImport but with an `as alias` suffix."""

  module: str

  def lines(self) -> List[str]:
    return [f"import {self.module} as {self.name}"]

  def sortkey(self) -> str:
    return self.module

  def change_alias(self, alias_name: str) -> ImportNode:
    return ImportAs(alias_name, self.module)


@dataclasses.dataclass(frozen=True)
class FromImportAs(ImportNode):
  """Same as FromImport but with an `as alias` suffix."""

  parent: str
  module: str

  def lines(self) -> List[str]:
    return [f"from {self.parent} import {self.module} as {self.name}"]

  def sortkey(self) -> str:
    return f"{self.parent}.{self.module}"

  def change_alias(self, alias_name: str) -> ImportNode:
    return FromImportAs(alias_name, self.parent, self.module)


@dataclasses.dataclass(frozen=True)
class Assignment(CodegenNode):
  """Assigns an expression to a left-hand-side expression."""
  lhs: str  # Dot-separate path.
  expr: str  # Expression to assign.

  def lines(self) -> List[str]:
    return [f"{self.lhs} = {self.expr}"]


@dataclasses.dataclass(frozen=True)
class ReturnStmt(CodegenNode):
  """Returns an expression."""
  expr: str  # Expression to return.

  def lines(self) -> List[str]:
    return [f"return {self.expr}"]


@dataclasses.dataclass(frozen=True)
class TrailingComment(CodegenNode):
  """Attaches a trailing comment to the last line of another node."""
  child: CodegenNode
  comment: str

  def lines(self) -> List[str]:
    comment = f"# {self.comment}"
    lines = self.child.lines()
    return [*lines[:-1], f"{lines[-1]}  {comment}"] if lines else [comment]


def block(sub_nodes_or_lines: List[Union[str, List[str], CodegenNode]],
          separator: List[str]) -> List[str]:
  """Helper to generate lines for a block of code.

  Args:
    sub_nodes_or_lines: Items to join, which are either string constants or
      codegen nodes.
    separator: Lines to insert between items to join.

  Returns:
    List of lines, taken from `sub_nodes_or_lines`.
  """
  sub_nodes_or_lines = filter(None, sub_nodes_or_lines)  # Remove empty items.

  result = []
  for i, item in enumerate(sub_nodes_or_lines):
    if i > 0:
      result.extend(separator)

    if isinstance(item, str):
      result.append(item)
    elif isinstance(item, list) and all(isinstance(x, str) for x in item):
      result.extend(item)
    elif isinstance(item, CodegenNode):
      result.extend(item.lines())
    else:
      raise TypeError(f"Unsupported item in argument to block: {item}")
  return result


@dataclasses.dataclass(frozen=True)
class ImmediateAttrsBlock(CodegenNode):
  """Block of code setting up immediate objects (as opposed to nested ones)."""
  nodes: List[CodegenNode]

  def lines(self) -> List[str]:
    return block(self.nodes, [])


@dataclasses.dataclass(frozen=True)
class SharedThenResultAssignment(CodegenNode):
  """Block of code setting shared instances, then a tree."""
  shared_instances: List[CodegenNode]
  tree_blocks: List[CodegenNode]

  def lines(self) -> List[str]:
    return block(self.shared_instances + self.tree_blocks, [""])


@dataclasses.dataclass(frozen=True)
class ConfigBuilder(CodegenNode):
  """Top-level node for imports, followed by builder code.

  By default, this generates a `build_config()` method.
  """
  imports: List[CodegenNode]
  builder_body: List[CodegenNode]

  def lines(self) -> List[str]:
    fn_body = ["  " + line for line in block(self.builder_body, [""])]
    return block([block(self.imports, []), ["def build_config():"] + fn_body],
                 ["", ""])



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for daglish."""

from typing import List

from absl.testing import absltest
import fiddle as fdl
from fiddle.experimental import daglish


class DaglishTest(absltest.TestCase):

  def test_path_fragment(self):
    cfg = fdl.Config(lambda x: x)
    path: List[daglish.PathElement] = [
        daglish.ListOrTupleItem([4], 1),
        daglish.DictItem({}, "a"),
        daglish.AttributeItem(object(), "foo"),
        daglish.DictItem({}, 2),
        daglish.AttributeItem(cfg, "bar"),
    ]
    path_str = "".join(x.code for x in path)
    self.assertEqual(path_str, "[1]['a'].foo[2].bar")


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A registry-based system for constructing models.

## Skeletons and fixtures

Skeletons generally express a tree of modules; for example in an EncoderLayer,
one might construct Flax modules like so:

EncoderLayer
  attention: normalized_block()
    layer_norm: layer_norm()
    body: self_attention()
      wrapped: MultiHeadDotProductAttention()
    dropout: dropout()
  mlp: normalized_block()
    layer_norm: layer_norm()
    body: mlp()
      wrapped: MlpBlock()
    dropout: dropout()

and this becomes even more detailed when libraries are composed. In order to
mitigate some verbosity, skeletons allow these tree-like templates to be
indexed by the type of the module.

However, details of configuration, typically specific to a project like T5,
should be filled in later. We call this configuration a "fixture". Generally
"skeletons" should only make choices that are quite natural to the types
themselves; for example, if an EncoderLayer has an "attention" and "mlp" fields,
then those are almost always filled with normalized blocks containing attention
and MLP modules.

## Validators

Validators are another utility to separate concerns between module authors and
projects which may use those modules. A module author might know that their
MlpBlock() module must have a parameter `foo` specified whenever another
parameter `bar` is True, and it is convenient to express those invariants at
the level of the type (MlpBlock, in this case) rather than counting on each
project / experimenter to reproduce those checks.

## Global vs. local registries

In general, we are going to have a global type-based registry for constructing
"skeletons" of models, and then use functions that set configuration in a nested
manner. However, the underlying implementation here is rather unopinionated in
terms of how "global" its state is, so if specific projects find non-global
registries more useful, they can easily work with them.

TBD:

 * Examples with placeholders.
 * Add a function for validating configs by traversing the config tree.
"""

import collections
import dataclasses
import inspect
from typing import Any, Callable, Dict, Generic, List, Optional, Type, TypeVar, Union

import fiddle as fdl
import typing_extensions

S = TypeVar("S")
T = TypeVar("T")

FnOrCls = Union[Type[S], Callable[..., S]]


class SkeletonFn(Generic[S], typing_extensions.Protocol):

  def __call__(self, cfg: fdl.Config[S]) -> None:
    pass


class ValidatorFn(Generic[S], typing_extensions.Protocol):

  def __call__(self, cfg: fdl.Config[S]) -> None:
    pass


def _is_configurable(fn_or_cls: FnOrCls) -> bool:
  """Helper method for determining wheher a function or class can be configured.

  This will return False for builtin stuff like int/tuple/etc. We generally
  don't want to configure these classes, partly because they work without
  requisite arguments, for example int() is 0.

  Args:
    fn_or_cls: Function or class.

  Returns:
    Whether `fn_or_cls` can be configured.
  """

  try:
    has_signature = inspect.signature(fn_or_cls) is not None
  except ValueError:
    has_signature = False

  return (has_signature  #
          and not inspect.isbuiltin(fn_or_cls)  #
          and not inspect.ismethod(fn_or_cls))


class DuplicateSkeletonError(ValueError):
  """Indicates that two skeletons were attempted to be registered for a type."""

  def __init__(self, fn_or_cls: FnOrCls[Any]):
    super().__init__(f"SkeletonFn for {fn_or_cls} already exists!")
    self.fn_or_cls = fn_or_cls

  def __str__(self):
    filename = inspect.getsourcefile(self.fn_or_cls)
    _, line_number = inspect.getsourcelines(self.fn_or_cls)
    return (f"SkeletonFn for {self.fn_or_cls} (defined at "
            f"{filename}:{line_number}) already exists!")


@dataclasses.dataclass
class TableEntry(Generic[T]):
  """An entry in the registry for a specific type or function.

  Attributes:
    skeleton_fn: A function which can create an initial configuration for this
      type or function.
    validators: A list of functions which can validate configuration objects of
      this type or function.
  """
  skeleton_fn: Optional[SkeletonFn[T]] = None
  validators: List[ValidatorFn[T]] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Registry:
  """An AutoBuilder registry, which knows how to create a model skeleton.

  Attributes:
    table: A mapping from a type or function to configure, to a TableEntry.
  """
  table: Dict[FnOrCls[Any], TableEntry[Any]] = dataclasses.field(
      default_factory=lambda: collections.defaultdict(TableEntry))

  def config(self,
             fn_or_cls: FnOrCls[T],
             require_skeleton: bool = True) -> fdl.Config[T]:
    """Creates a configuration instance of a given function or class.

    Args:
      fn_or_cls: Function or class to get a skeleton fdl.Config instance for.
      require_skeleton: If True, a KeyError will be raised if a skeleton of the
        desired function/type is not available. If False, an empty fdl.Config of
        that function/class will be returned.

    Returns:
      Configuration instance.

    Raises:
      KeyError: If there is no skeleton registered for the given class, and
        require_skeleton is True (the default).
    """
    base_config = fdl.Config(fn_or_cls)
    entry = self.table[fn_or_cls]
    if entry.skeleton_fn is None:
      if require_skeleton:
        raise KeyError(f"No skeleton registered for {fn_or_cls}.")
      else:
        return base_config
    entry.skeleton_fn(base_config)
    return base_config

  def skeleton(
      self, fn_or_cls: FnOrCls[T]) -> Callable[[SkeletonFn[T]], SkeletonFn[T]]:
    """Registers a function as a skeleton for a given type.

    Example:

    @dataclasses.dataclass
    class MyDense:
      activation: Callable[..., Any]

    @ab.skeleton(MyDense):
    def my_dense_skeleton(config: fdl.Config):
      config.activation = fdl.Config(nn.gelu, approximate=True)

    Often the skeleton definition (`my_dense_skeleton`) can be kept separately
    from the modeling library code, to make the modeling library code agnostic
    to the choice of configuration library.

    Args:
      fn_or_cls: Function that the skeleton is configuring.

    Returns:
      Decorator that, when applied to a function definition, will register that
      function as a skeleton.
    """

    def inner(skeleton_fn):
      entry = self.table[fn_or_cls]
      if entry.skeleton_fn is not None:
        raise DuplicateSkeletonError(fn_or_cls)
      entry.skeleton_fn = skeleton_fn
      return skeleton_fn

    return inner

  def auto_skeleton(self,
                    fn_or_cls: FnOrCls[Any],
                    attributes_require_skeletons: bool = False,
                    **kwargs) -> None:
    """Registers a skeleton based on arg types.

    Sometimes functions are written to take a more generic class, for example
    one might have an encoder layer in Flax expressed like,

    class EncoderLayer(nn.Module):
      dropout: nn.Module

    but in practice, `dropout` is almost always `nn.Dropout`, the type signature
    is just left to be the most general thing in case of experimental overrides.

    In this case, we can specify that the dropout is `nn.Dropout` as follows:

    ab.auto_skeleton(EncoderLayer, dropout=nn.Dropout)

    This is equivalent to writing:

    @ab.skeleton(EncoderLayer)
    def encoder_layer_skeleton(config: fdl.Config[EncoderLayer]):
      config.dropout = ab.config(nn.Dropout, require_skeleton=False)

    By default, `auto_skeleton` doesn't require the attributes to have defined
    skeletons, but would use them if available. In this case, the `dropout`
    class attribute can be filled with a blank `fdl.Config(nn.Dropout)`, if
    there is not a skeleton for `nn.Dropout`.

    If you require more advanced skeleton setups, then please use the `skeleton`
    method.

    Args:
      fn_or_cls: Function or class to generate the skeleton for.
      attributes_require_skeletons: Whether skeletons are required for any class
        attributes or function arguments to `fn_or_cls`. By default, skeletons
        are not required, whereupon empty fdl.Config instances are created.
      **kwargs: Any override types to use for configuration.

    Raises:
      ValueError: If there is no annotation for a parameter not mentioned in
        **kwargs.
      TypeError: If one of the parameters not mentioned in **kwargs is not a
        configurable type.
    """
    arg_to_type = {}
    for name, parameter in inspect.signature(fn_or_cls).parameters.items():
      if name in kwargs:
        # None is a special value that indicates no type should be configured
        # for the parameter.
        if kwargs[name] is not None:
          arg_to_type[name] = kwargs[name]
      elif parameter.annotation == inspect.Signature.empty:
        raise ValueError(
            f"Parameter {name!r} of {fn_or_cls} doesn't have an annotation, "
            "and no override type was set in the auto_skeleton decorator.")
      elif not _is_configurable(parameter.annotation):
        raise TypeError(
            f"Parameter {name!r} of {fn_or_cls} is of type "
            f"{parameter.annotation}, which cannot be configured. Please pass "
            f"{name}=None to your auto-skeleton invocation to avoid "
            "automatically adding a Config for this parameter.")
      else:
        arg_to_type[name] = parameter.annotation

    def arg_type_skeleton(cfg: fdl.Config):
      for name, type_to_configure in arg_to_type.items():
        setattr(
            cfg, name,
            self.config(
                type_to_configure,
                require_skeleton=attributes_require_skeletons))

    self.skeleton(fn_or_cls)(arg_type_skeleton)

  def validator(
      self,
      fn_or_cls: FnOrCls[T]) -> Callable[[ValidatorFn[T]], ValidatorFn[T]]:
    """Decorator to register a function as a validator for a given type.

    Args:
      fn_or_cls: Function that the validator is checking.

    Returns:
      Decorator that, when applied to a function definition, will register that
      function as a validator.
    """

    def inner(validator_fn: ValidatorFn[T]) -> ValidatorFn[T]:
      self.table[fn_or_cls].validators.append(validator_fn)
      return validator_fn

    return inner


# By default, the global instance should be used. Its methods are aliased to the
# module level.
_default_registry = Registry()
config = _default_registry.config
skeleton = _default_registry.skeleton
auto_skeleton = _default_registry.auto_skeleton
validator = _default_registry.validator



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The public API for the absl_flags Fiddle extension."""

from fiddle.absl_flags.flags import apply_fiddlers_to
from fiddle.absl_flags.flags import apply_overrides_to
from fiddle.absl_flags.flags import create_buildable_from_flags
from fiddle.absl_flags.flags import flags_parser



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides `autofill` to instantiate child configs for configurable arguments.

When configuring a type that depends on many other child types that are
themselves configurable, nested `fdl.Config` instances must be constructed for
each of them. This is can involve large amounts of mechanical boilerplate.

When a function or class has type annotations for its arguments, `autofill`
will recursively instantiate `fdl.Config`s for them.

`autofill` has the added benefit of further incentivizing type annotations, too!

By convention, `autofill` is often imported `as af` (which can also be
pronounced "auto-Fiddle").
"""

import dataclasses
from typing import Any, Type

from fiddle import config


def _is_autofillable(annotation: Type[Any]) -> bool:
  """A heuristic for when `fdl.Config`'s should be instantiated.

  This heuristic guesses that a nested `fdl.Config` should be instantiated
  if the type annotation corresponds to a dataclass or if the type has a
  `__init_fiddle__` defined.

  Args:
    annotation: The type annotated to an argument to a Buildable.

  Returns:
    True if the argument corresponding to this annotation should be autofilled,
    False otherwise.
  """
  return (dataclasses.is_dataclass(annotation) or
          hasattr(annotation, '__fiddle_init__'))


def autofill(cfg: config.Buildable):
  """Recursively instantiates `fdl.Config`'s for arguments.

  For any argument that does not already have an assigned value, an autofilled
  `fdl.Config` will be assigned if the argument's type annotation is either a
  dataclass or has a `__init_fiddle__` method. Arguments without a type
  annotation are never autofilled.

  Args:
    cfg: The root of the configuration to fill out. This Buildable is modified
      in-place.
  """
  # TODO: Explore adding a configurable policy argument.
  for param in cfg.__signature__.parameters.values():

    if param.name in cfg.__arguments__:
      # Param already has a value.
      continue

    if param.annotation is param.empty:
      # No annotation; can't instantiate a config if you don't know the type!
      continue

    if param.default is not param.empty:
      # Param has a default; will just take that.
      continue

    if param.kind not in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY):
      # If the parameter is variable (vararg, varkwarg) or is positional-only,
      # it is not supported.
      continue

    if not _is_autofillable(param.annotation):
      # Policy doesn't allow for autobuilding.
      continue

    child = config.Config(param.annotation)
    autofill(child)
    setattr(cfg, param.name, child)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Init file for the `fiddle` package."""

from fiddle.building import build
from fiddle.building import BuildError
from fiddle.config import assign
from fiddle.config import Buildable
from fiddle.config import Config
from fiddle.config import Partial
from fiddle.config import update_callable
from fiddle.materialize import materialize_defaults
from fiddle.tagging import set_tagged
from fiddle.tagging import Tag
from fiddle.tagging import TaggedValue
